# -*- coding: utf-8 -*-
"""
Created on Tue May 11 15:31:31 2021

:copyright: 
    Jared Peacock (jpeacock@usgs.gov)

:license: MIT

"""
# =============================================================================
# Imports
# =============================================================================
from pathlib import Path
from io import StringIO
import warnings
from copy import deepcopy

# supress the future warning from pandas about using datetime parser.
warnings.simplefilter(action="ignore", category=FutureWarning)

import pandas as pd
import logging
from datetime import datetime

from mth5.timeseries import ChannelTS, RunTS
from mt_metadata.timeseries import Station, Run, Electric, Magnetic, Auxiliary
from mt_metadata.utils.mttime import MTime

# =============================================================================
def lemi_date_parser(year, month, day, hour, minute, second):
    """
    convenience function to combine the date-time columns that are output by
    lemi into a single column

    Assumes UTC

    :param year: year
    :type year: int
    :param month: month
    :type month: int
    :param day: day of the month
    :type day: int
    :param hour: hour in 24 hr format
    :type hour: int
    :param minute: minutes in the hour
    :type minute: int
    :param second: seconds in the minute
    :type second: int
    :return: date time as a single column
    :rtype: :class:`pandas.DateTime`

    """

    return pd.to_datetime(
        datetime(
            int(year),
            int(month),
            int(day),
            int(hour),
            int(minute),
            int(second),
        )
    )


def lemi_position_parser(position):
    """
    convenience function to parse the location strings into a decimal float
    Uses the hemisphere for the sign.

    .. note:: the format of the location is odd in that it is multiplied by
     100 within the LEMI to provide a single floating point value that
     includes the degrees and decimal degrees --> {degrees}{degrees[mm.ss]}.
     For example 40.50166 would be represented as 4030.1.

    :param position: DESCRIPTION
    :type position: TYPE
    :return: DESCRIPTION
    :rtype: TYPE

    """
    pos = f"{float(position) / 100}".split(".")
    degrees = int(pos[0])
    decimals = float(f"{pos[1][0:2]}.{pos[1][2:]}") / 60

    location = degrees + decimals
    return location


def lemi_hemisphere_parser(hemisphere):
    """
    convert hemisphere into a value [-1, 1].  Assumes the prime meridian is 0.

    :param hemisphere: hemisphere string [ 'N' | 'S' | 'E' | 'W']
    :type hemisphere: string
    :return: unity with a sign for the given hemisphere
    :rtype: signed integer

    """
    if hemisphere in ["S", "W"]:
        return -1
    return 1


class LEMI424:
    """
    Read in a LEMI424 file, this is a place holder until IRIS finalizes
    their reader.

    :param fn: full path to LEMI424 file
    :type fn: :class:`pathlib.Path` or string
    :param sample_rate: sample rate of the file, default is 1.0
    :type sample_rate: float
    :param chunk_size: chunk size for pandas to use, does not change reading
     time much for a single day file. default is 8640
    :type chunk_size: integer
    :param file_column_names: column names of the LEMI424 file
    :type file_column_names: list of strings
    :param dtypes: data types for each column
    :type dtypes: dictionary with keys of column names and values of data types
    :param data_column_names: same as file_column names with and added column
     for date, which is the combined date and time columns.
    :type data_column_names: dictionary with keys of column names and values
     of data types

    :LEMI424 File Column Names:

        - **year**
        - **month**
        - **day**
        - **hour**
        - **minute**
        - **second**
        - **bx**
        - **by**
        - **bz**
        - **temperature_e**
        - **temperature_h**
        - **e1**
        - **e2**
        - **e3**
        - **e4**
        - **battery**
        - **elevation**
        - **latitude**
        - **lat_hemisphere**
        - **longitude**
        - **lon_hemisphere**
        - **n_satellites**
        - **gps_fix**
        - **time_diff**

    :Data Column Names:

        - **date**
        - **bx**
        - **by**
        - **bz**
        - **temperature_e**
        - **temperature_h**
        - **e1**
        - **e2**
        - **e3**
        - **e4**
        - **battery**
        - **elevation**
        - **latitude**
        - **lat_hemisphere**
        - **longitude**
        - **lon_hemisphere**
        - **n_satellites**
        - **gps_fix**
        - **time_diff**

    """

    def __init__(self, fn=None, **kwargs):
        self.logger = logging.getLogger(f"{__name__}.{self.__class__.__name__}")
        self.fn = fn
        self.sample_rate = 1.0
        self.chunk_size = 8640
        self.data = None
        self.file_column_names = [
            "year",
            "month",
            "day",
            "hour",
            "minute",
            "second",
            "bx",
            "by",
            "bz",
            "temperature_e",
            "temperature_h",
            "e1",
            "e2",
            "e3",
            "e4",
            "battery",
            "elevation",
            "latitude",
            "lat_hemisphere",
            "longitude",
            "lon_hemisphere",
            "n_satellites",
            "gps_fix",
            "time_diff",
        ]

        self.dtypes = dict(
            [
                ("year", int),
                ("month", int),
                ("day", int),
                ("hour", int),
                ("minute", int),
                ("second", int),
                ("bx", float),
                ("by", float),
                ("bz", float),
                ("temperature_e", float),
                ("temperature_h", float),
                ("e1", float),
                ("e2", float),
                ("e3", float),
                ("e4", float),
                ("battery", float),
                ("elevation", float),
                ("n_satellites", int),
                ("gps_fix", int),
                ("time_diff", float),
            ]
        )

        self.data_column_names = ["date"] + self.file_column_names[6:]

    def __str__(self):
        lines = ["LEMI 424 data", "-" * 20]
        lines.append(f"start:      {self.start.isoformat()}")
        lines.append(f"end:        {self.end.isoformat()}")
        lines.append(f"N samples:  {self.n_samples}")
        lines.append(f"latitude:   {self.latitude} (degrees)")
        lines.append(f"longitude:  {self.longitude} (degrees)")
        lines.append(f"elevation:  {self.elevation} m")

        return "\n".join(lines)

    def __repr__(self):
        return self.__str__()

    def __add__(self, other):
        """
        Can append other LEMI424 objects together as long as the start and
        end times match up.

        """
        if not self._has_data():
            raise ValueError("Data is None cannot append to. Read file first")

        if isinstance(other, LEMI424):
            new = LEMI424()
            new.__dict__.update(self.__dict__)
            new.data = pd.concat([new.data, other.data])
            return new

        elif isinstance(other, pd.DataFrame):

            if not other.columns != self.data.columns:
                raise ValueError("DataFrame columns are not the same.")

            new = LEMI424()
            new.__dict__.update(self.__dict__)
            new.data = pd.concat([new.data, other])
            return new

        else:
            raise ValueError(f"Cannot add {type(other)} to pd.DataFrame.")

    @property
    def data(self):
        """
        Data represented as a :class:`pandas.DataFrame` with data_column names

        """
        return self._data

    @data.setter
    def data(self, data):
        """
        Make sure if data is set that it is a pandas data frame with the proper
        column names
        """

        if data is None:
            self._data = None

        elif isinstance(data, pd.DataFrame):
            if len(data.columns) == len(self.file_column_names):
                if data.columns != self.file_column_names:
                    raise ValueError(
                        "Column names are not the same. "
                        "Check LEMI424.column_names for accepted names"
                    )
            elif len(data.columns) == len(self.data_column_names):
                if not data.columns == self.data_column_names:
                    raise ValueError(
                        "Column names are not the same. "
                        "Check LEMI424.column_names for accepted names"
                    )
            else:
                self._data = data

    def _has_data(self):
        """check to see if has data or not"""
        if self.data is not None:
            return True
        return False

    @property
    def fn(self):
        """full path to LEMI424 file"""
        return self._fn

    @fn.setter
    def fn(self, value):
        if value is not None:
            value = Path(value)
            if not value.exists():
                raise IOError(f"Could not find {value}")
        self._fn = value

    @property
    def file_size(self):
        """size of file in bytes"""
        if self.fn is not None:
            return self.fn.stat().st_size

    @property
    def start(self):
        """start time of data collection in the LEMI424 file"""
        if self._has_data():
            return MTime(self.data.index[0])

    @property
    def end(self):
        """end time of data collection in the LEMI424 file"""
        if self._has_data():
            return MTime(self.data.index[-1])

    @property
    def latitude(self):
        """median latitude where data have been collected in the LEMI424 file"""
        if self._has_data():

            return (
                self.data.latitude.median() * self.data.lat_hemisphere.median()
            )

    @property
    def longitude(self):
        """median longitude where data have been collected in the LEMI424 file"""
        if self._has_data():
            return (
                self.data.longitude.median() * self.data.lon_hemisphere.median()
            )

    @property
    def elevation(self):
        """median elevation where data have been collected in the LEMI424 file"""
        if self._has_data():
            return self.data.elevation.median()

    @property
    def n_samples(self):
        """number of samples in the file"""
        if self._has_data():
            return self.data.shape[0]
        elif self.fn is not None and self.fn.exists():
            return round(self.fn.stat().st_size / 152.0)

    @property
    def gps_lock(self):
        """has GPS lock"""
        if self._has_data():
            return self.data.gps_fix.values

    @property
    def station_metadata(self):
        """station metadata as :class:`mt_metadata.timeseries.Station`"""
        s = Station()
        if self._has_data():
            s.location.latitude = self.latitude
            s.location.longitude = self.longitude
            s.location.elevation = self.elevation
            s.time_period.start = self.start
            s.time_period.end = self.end
            s.add_run(self.run_metadata)
        return s

    @property
    def run_metadata(self):
        """run metadata as :class:`mt_metadata.timeseries.Run`"""
        r = Run()
        r.id = "a"
        r.sample_rate = self.sample_rate
        r.data_logger.model = "LEMI424"
        r.data_logger.manufacturer = "LEMI"
        if self._has_data():
            r.data_logger.power_source.voltage.start = self.data.battery.max()
            r.data_logger.power_source.voltage.end = self.data.battery.min()
            r.time_period.start = self.start
            r.time_period.end = self.end

            for ch_aux in ["temperature_e", "temperature_h"]:
                r.add_channel(Auxiliary(component=ch_aux))
            for ch_e in ["e1", "e2"]:
                r.add_channel(Electric(component=ch_e))
            for ch_h in ["bx", "by", "bz"]:
                r.add_channel(Magnetic(component=ch_h))

        return r

    def read(self, fn=None, fast=True):
        """
        Read a LEMI424 file using pandas.  The `fast` way will read in the
        first and last line to get the start and end time to make a time index.
        Then it will read in the data skipping parsing the date time columns.
        It will check to make sure the expected amount of points are correct.
        If not then it will read in the slower way which used the date time
        parser to ensure any time gaps are respected.

        :param fn: full path to file, defaults to None.  Uses LEMI424.fn if
         not provided
        :type fn: string or :class:`pathlib.Path`, optional
        :param fast: read the fast way (True) or not (False)
        :return: DESCRIPTION
        :rtype: TYPE

        """
        if fn is not None:
            self.fn = fn

        if not self.fn.exists():
            msg = "Could not find file %s"
            self.logger.error(msg, self.fn)
            raise IOError(msg % self.fn)

        if fast:
            try:
                self.read_metadata()
                data = pd.read_csv(
                    self.fn,
                    delimiter="\s+",
                    names=self.file_column_names,
                    dtype=self.dtypes,
                    usecols=(
                        "bx",
                        "by",
                        "bz",
                        "temperature_e",
                        "temperature_h",
                        "e1",
                        "e2",
                        "e3",
                        "e4",
                        "battery",
                        "elevation",
                        "latitude",
                        "lat_hemisphere",
                        "longitude",
                        "lon_hemisphere",
                        "n_satellites",
                        "gps_fix",
                        "time_diff",
                    ),
                    converters={
                        "latitude": lemi_position_parser,
                        "longitude": lemi_position_parser,
                        "lat_hemisphere": lemi_hemisphere_parser,
                        "lon_hemisphere": lemi_hemisphere_parser,
                    },
                )
                time_index = pd.date_range(
                    start=self.start.iso_no_tz,
                    end=self.end.iso_no_tz,
                    freq="1000000000N",
                )
                if time_index.size != data.shape[0]:
                    raise ValueError(
                        "Missing a time stamp use read with fast=False"
                    )

                data.index = time_index
                self.data = data
                return
            except ValueError:
                self.logger.warning(
                    "Data is missing a time stamp, reading in slow mode"
                )

        # read in chunks, this doesnt really speed up much as most of the
        # compute time is used in the date time parsing.
        if self.n_samples > self.chunk_size:
            st = MTime().now()
            dfs = list(
                pd.read_csv(
                    self.fn,
                    delimiter="\s+",
                    names=self.file_column_names,
                    dtype=self.dtypes,
                    parse_dates={
                        "date": [
                            "year",
                            "month",
                            "day",
                            "hour",
                            "minute",
                            "second",
                        ]
                    },
                    date_parser=lemi_date_parser,
                    converters={
                        "latitude": lemi_position_parser,
                        "longitude": lemi_position_parser,
                        "lat_hemisphere": lemi_hemisphere_parser,
                        "lon_hemisphere": lemi_hemisphere_parser,
                    },
                    index_col="date",
                    chunksize=self.chunk_size,
                )
            )

            self.data = pd.concat(dfs)
            et = MTime().now()
            self.logger.debug(
                f"Reading {self.fn.name} took {et - st:.2f} seconds"
            )

        else:
            st = MTime().now()
            self.data = pd.read_csv(
                self.fn,
                delimiter="\s+",
                names=self.file_column_names,
                dtype=self.dtypes,
                parse_dates={
                    "date": [
                        "year",
                        "month",
                        "day",
                        "hour",
                        "minute",
                        "second",
                    ]
                },
                date_parser=lemi_date_parser,
                converters={
                    "latitude": lemi_position_parser,
                    "longitude": lemi_position_parser,
                    "lat_hemisphere": lemi_hemisphere_parser,
                    "lon_hemisphere": lemi_hemisphere_parser,
                },
                index_col="date",
            )
            et = MTime().now()
            self.logger.debug(
                f"Reading {self.fn.name} took {et - st:.2f} seconds"
            )

    def read_metadata(self):
        """
        Read only first and last rows to get important metadata to use in
        the collection.

        """

        with open(self.fn) as fid:
            first_line = fid.readline()
            for line in fid:
                pass  # iterate to the end
            last_line = line

        lines = StringIO(f"{first_line}\n{last_line}")

        self.data = pd.read_csv(
            lines,
            delimiter="\s+",
            names=self.file_column_names,
            dtype=self.dtypes,
            parse_dates={
                "date": [
                    "year",
                    "month",
                    "day",
                    "hour",
                    "minute",
                    "second",
                ]
            },
            date_parser=lemi_date_parser,
            converters={
                "latitude": lemi_position_parser,
                "longitude": lemi_position_parser,
                "lat_hemisphere": lemi_hemisphere_parser,
                "lon_hemisphere": lemi_hemisphere_parser,
            },
            index_col="date",
        )

    def to_run_ts(self, fn=None, e_channels=["e1", "e2"]):
        """
        Create a :class:`mth5.timeseries.RunTS` object from the data

        :param fn: full path to file, defaults to None.  Will use LEMI424.fn
         if None.
        :type fn: string or :class:`pathlib.Path`, optional
        :param e_channels: columns for the electric channels to use,
         defaults to ["e1", "e2"]
        :type e_channels: list of strings, optional
        :return: RunTS object
        :rtype: :class:`mth5.timeseries.RunTS`

        """

        ch_list = []

        for comp in (
            ["bx", "by", "bz"] + e_channels + ["temperature_e", "temperature_h"]
        ):
            if comp[0] in ["h", "b"]:
                ch = ChannelTS("magnetic")

            elif comp[0] in ["e"]:
                ch = ChannelTS("electric")
            else:
                ch = ChannelTS("auxiliary")

            ch.sample_rate = self.sample_rate
            ch.start = self.start
            ch.ts = self.data[comp].values
            ch.component = comp

            ch_list.append(ch)

        return RunTS(
            array_list=ch_list,
            station_metadata=self.station_metadata,
            run_metadata=self.run_metadata,
        )


# =============================================================================
# define the reader
# =============================================================================
def read_lemi424(
    fn, e_channels=["e1", "e2"], fast=True, logger_file_handler=None
):
    """
    Read a LEMI 424 TXT file.

    :param fn: input file name
    :type fn: string or Path
    :param e_channels: A list of electric channels to read,
    defaults to ["e1", "e2"]
    :type e_channels: list of strings, optional
    :return: A RunTS object with appropriate metadata
    :rtype: :class:`mth5.timeseries.RunTS`

    """

    if not isinstance(fn, (list, tuple)):
        fn = [fn]

    txt_obj = LEMI424(fn[0])
    txt_obj.read(fast=fast)

    if logger_file_handler:
        txt_obj.logger.addHandler(logger_file_handler)

    # read a list of files into a single run
    if len(fn) > 1:
        for txt_file in fn:
            other = LEMI424(txt_file)
            other.read()
            txt_obj += other

    return txt_obj.to_run_ts(e_channels=e_channels)



# -*- coding: utf-8 -*-
"""
Test time series

Created on Tue Jun 30 16:38:27 2020

:copyright:
    author: Jared Peacock
    
:license:
    MIT
    
"""
# =============================================================================
# imports
# =============================================================================

import unittest

import numpy as np

from mth5.timeseries import ChannelTS, RunTS
from mth5.utils.exceptions import MTTSError

from mt_metadata.utils.mttime import MTime
import mt_metadata.timeseries as metadata
from mt_metadata.timeseries.filters import (
    PoleZeroFilter,
    ChannelResponseFilter,
)

# =============================================================================
# test run
# =============================================================================
class TestRunTS(unittest.TestCase):
    def setUp(self):
        self.run = RunTS()
        self.maxDiff = None
        self.start = "2015-01-08T19:49:18+00:00"
        self.end = "2015-01-08T19:57:49.875000"
        self.sample_rate = 8
        self.npts = 4096

        pz = PoleZeroFilter(
            units_in="volts", units_out="nanotesla", name="instrument_response"
        )
        pz.poles = [
            (-6.283185 + 10.882477j),
            (-6.283185 - 10.882477j),
            (-12.566371 + 0j),
        ]
        pz.zeros = []
        pz.normalization_factor = 18244400

        self.cr = ChannelResponseFilter(filters_list=[pz])

        self.ex = ChannelTS(
            "electric",
            data=np.random.rand(self.npts),
            channel_metadata={
                "electric": {
                    "component": "Ex",
                    "sample_rate": self.sample_rate,
                    "time_period.start": self.start,
                }
            },
            channel_response_filter=self.cr,
        )
        self.ey = ChannelTS(
            "electric",
            data=np.random.rand(self.npts),
            channel_metadata={
                "electric": {
                    "component": "Ey",
                    "sample_rate": self.sample_rate,
                    "time_period.start": self.start,
                }
            },
            channel_response_filter=self.cr,
        )
        self.hx = ChannelTS(
            "magnetic",
            data=np.random.rand(self.npts),
            channel_metadata={
                "magnetic": {
                    "component": "hx",
                    "sample_rate": self.sample_rate,
                    "time_period.start": self.start,
                }
            },
            channel_response_filter=self.cr,
        )
        self.hy = ChannelTS(
            "magnetic",
            data=np.random.rand(self.npts),
            channel_metadata={
                "magnetic": {
                    "component": "hy",
                    "sample_rate": self.sample_rate,
                    "time_period.start": self.start,
                }
            },
            channel_response_filter=self.cr,
        )
        self.hz = ChannelTS(
            "magnetic",
            data=np.random.rand(self.npts),
            channel_metadata={
                "magnetic": {
                    "component": "hz",
                    "sample_rate": self.sample_rate,
                    "time_period.start": self.start,
                }
            },
            channel_response_filter=self.cr,
        )

        self.run.set_dataset([self.ex, self.ey, self.hx, self.hy, self.hz])

    def test_str(self):
        s_list = [
            f"Survey:      {self.run.survey_metadata.id}",
            f"Station:     {self.run.station_metadata.id}",
            f"Run:         {self.run.run_metadata.id}",
            f"Start:       {self.run.start}",
            f"End:         {self.run.end}",
            f"Sample Rate: {self.run.sample_rate}",
            f"Components:  {self.run.channels}",
        ]
        test_str = "\n\t".join(["RunTS Summary:"] + s_list)

        self.assertEqual(test_str, self.run.__str__())

    def test_repr(self):
        s_list = [
            f"Survey:      {self.run.survey_metadata.id}",
            f"Station:     {self.run.station_metadata.id}",
            f"Run:         {self.run.run_metadata.id}",
            f"Start:       {self.run.start}",
            f"End:         {self.run.end}",
            f"Sample Rate: {self.run.sample_rate}",
            f"Components:  {self.run.channels}",
        ]
        test_str = "\n\t".join(["RunTS Summary:"] + s_list)

        self.assertEqual(test_str, self.run.__repr__())

    def test_set_run_metadata_fail(self):
        self.assertRaises(TypeError, RunTS, [self.ex], **{"run_metadata": []})

    def test_set_station_metadata_fail(self):
        self.assertRaises(
            TypeError, RunTS, [self.ex], **{"station_metadata": []}
        )

    def test_validate_run_metadata(self):
        self.assertEqual(
            self.run.run_metadata,
            self.run._validate_run_metadata(self.run.run_metadata),
        )

    def test_validate_run_metadata_from_dict(self):
        self.assertEqual(
            metadata.Run(id="0"),
            self.run._validate_run_metadata({"id": "0"}),
        )

    def test_validate_station_metadata(self):
        self.assertEqual(
            self.run.station_metadata,
            self.run._validate_station_metadata(self.run.station_metadata),
        )

    def test_validate_station_metadata_from_dict(self):
        self.assertEqual(
            metadata.Station(id="0"),
            self.run._validate_station_metadata({"id": "0"}),
        )

    def test_validate_survey_metadata(self):
        self.assertEqual(
            self.run.survey_metadata,
            self.run._validate_survey_metadata(self.run.survey_metadata),
        )

    def test_validate_survey_metadata_from_dict(self):
        self.assertEqual(
            metadata.Survey(id="0"),
            self.run._validate_survey_metadata({"id": "0"}),
        )

    def test_validate_array_fail(self):
        with self.subTest("bad type"):
            self.assertRaises(TypeError, self.run._validate_array_list, 10)

        with self.subTest("bad list"):
            self.assertRaises(TypeError, self.run._validate_array_list, [10])

    def test_initialize(self):

        with self.subTest("channels"):
            self.assertListEqual(
                ["ex", "ey", "hx", "hy", "hz"], self.run.channels
            )
        with self.subTest("sample rate"):
            self.assertEqual(self.run.sample_rate, self.sample_rate)
        with self.subTest("start"):
            self.assertEqual(self.run.start, MTime(self.start))
        with self.subTest("end"):
            self.assertEqual(self.run.end, MTime(self.end))

    def test_sample_interval(self):
        self.assertEqual(1.0 / self.sample_rate, self.run.sample_interval)

    def test_sr_fail(self):
        hz = ChannelTS(
            "magnetic",
            data=np.random.rand(self.npts),
            channel_metadata={
                "magnetic": {
                    "component": "hz",
                    "sample_rate": 1,
                    "time_period.start": self.start,
                }
            },
        )

        self.assertRaises(
            ValueError,
            self.run.set_dataset,
            [self.ex, self.ey, self.hx, self.hy, hz],
        )

    def test_channels(self):

        for comp in ["ex", "ey", "hx", "hy", "hz"]:
            ch = getattr(self.run, comp)

            with self.subTest(msg=f"{comp} isinstance channel"):
                self.assertIsInstance(ch, ChannelTS)
            with self.subTest(msg=f"{comp} sample rate"):
                self.assertEqual(ch.sample_rate, self.sample_rate)
            with self.subTest(msg=f"{comp} start"):
                self.assertEqual(ch.start, MTime(self.start))
            with self.subTest(msg=f"{comp} end"):
                self.assertEqual(ch.end, MTime(self.end))
            with self.subTest(msg=f"{comp} component"):
                self.assertEqual(ch.component, comp)

            with self.subTest(msg=f"{comp} filters"):
                self.assertListEqual(
                    self.cr.filters_list,
                    ch.channel_response_filter.filters_list,
                )

    def test_get_channel_fail(self):
        """
        self.run.temperature should return None, because 'temperature' is not in self.channels
        :return:
        """

        self.assertRaises(NameError, getattr, *(self.run, "temperature"))

    def test_wrong_metadata(self):
        self.run.run_metadata.sample_rate = 10
        self.run.validate_metadata()

        with self.subTest("sample rate"):
            self.assertEqual(
                self.ex.sample_rate, self.run.run_metadata.sample_rate
            )
        with self.subTest("start"):
            self.run.run_metadata.start = "2020-01-01T00:00:00"
            self.run.validate_metadata()
            self.assertEqual(
                self.run.start, self.run.run_metadata.time_period.start
            )
        with self.subTest("end"):
            self.run.run_metadata.end = "2020-01-01T00:00:00"
            self.run.validate_metadata()
            self.assertEqual(
                self.run.end, self.run.run_metadata.time_period.end
            )

    def test_get_slice(self):

        start = "2015-01-08T19:49:30+00:00"
        npts = 256

        r_slice = self.run.get_slice(start, n_samples=npts)

        with self.subTest("isinstance runts"):
            self.assertIsInstance(r_slice, RunTS)
        with self.subTest("sample rate"):
            self.assertEqual(r_slice.sample_rate, self.sample_rate)
        with self.subTest("start not equal"):
            self.assertEqual(r_slice.start, MTime(start))

        with self.subTest("start equal"):
            # the time index does not have a value at the requested location
            # so it grabs the closest one.
            self.assertEqual(r_slice.start, MTime(start))
        with self.subTest("end"):
            self.assertEqual(
                r_slice.end, MTime("2015-01-08T19:50:01.875000+00:00")
            )

        with self.subTest("npts"):
            self.assertEqual(r_slice.dataset.ex.data.shape[0], npts)

    def test_filters_dict(self):
        self.assertEqual(list(self.run.filters.keys()), ["instrument_response"])

    def test_filters_fail(self):
        def set_filters(value):
            self.run.filters = value

        self.assertRaises(TypeError, set_filters, ())

    def test_summarize_metadata(self):
        meta_dict = {}
        for comp in self.run.dataset.data_vars:
            for mkey, mvalue in self.run.dataset[comp].attrs.items():
                meta_dict[f"{comp}.{mkey}"] = mvalue
        self.assertDictEqual(meta_dict, self.run.summarize_metadata)

    def test_add_channel_xarray(self):
        x = self.ex.to_xarray()
        x.attrs["component"] = "ez"
        x.name = "ez"
        self.run.add_channel(x)

        self.assertEquals(
            sorted(self.run.channels),
            sorted(["ex", "ey", "ez", "hx", "hy", "hz"]),
        )

    def test_to_obspy_stream(self):
        stream = self.run.to_obspy_stream()

        with self.subTest("count"):
            self.assertEqual(stream.count(), 5)

        for tr in stream.traces:
            with self.subTest("sample_rate"):
                self.assertEqual(tr.stats.sampling_rate, self.sample_rate)

            with self.subTest("start time"):
                self.assertEqual(tr.stats.starttime, self.start)

            with self.subTest("npts"):
                self.assertEqual(tr.stats.npts, self.npts)


# =============================================================================
# run tests
# =============================================================================
if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""
Helper functions for HDF5

Created on Tue Jun  2 12:37:50 2020

:copyright:
    Jared Peacock (jpeacock@usgs.gov)
    
:license: 
    MIT

"""
# =============================================================================
# Imports
# =============================================================================
from collections.abc import Iterable
import inspect
import numpy as np
import h5py
import gc

from mth5.utils.mth5_logger import setup_logger

logger = setup_logger(__name__)

# =============================================================================
# Acceptable compressions
# =============================================================================
COMPRESSION = ["lzf", "gzip", "szip", None]
COMPRESSION_LEVELS = {
    "lzf": [None],
    "gzip": range(10),
    "szip": ["ec-8", "ee-10", "nn-8", "nn-10"],
    None: [None],
}


def validate_compression(compression, level):
    """
    validate that the input compression is supported.

    :param compression: type of lossless compression
    :type compression: string, [ 'lzf' | 'gzip' | 'szip' | None ]
    :param level: compression level if supported
    :type level: string for 'szip' or int for 'gzip'
    :return: compression type
    :rtype: string
    :return: compressiong level
    :rtype: string for 'szip' or int for 'gzip'
    :raises: ValueError if comporession or level are not supported
    :raises: TypeError if compression level is not a string

    """
    if compression is None:
        return None, None
    if not isinstance(compression, (str, type(None))):
        msg = "compression type must be a string, not {0}".format(
            type(compression)
        )
        logger.error(msg)
        raise TypeError(msg)
    if not compression in COMPRESSION:
        msg = (
            f"Compression type {compression} not supported. "
            + f"Supported options are {COMPRESSION}"
        )
        logger.error(msg)
        raise ValueError(msg)
    if compression == "lzf":
        level = COMPRESSION_LEVELS["lzf"][0]
    elif compression == " gzip":
        if not isinstance(level, (int)):
            msg = "Level type for gzip must be an int, not {0}.".format(
                type(level)
                + f" Options are {0}".format(COMPRESSION_LEVELS["gzip"])
            )
            logger.error(msg)
            raise TypeError(msg)
    elif compression == " szip":
        if not isinstance(level, (str)):
            msg = "Level type for szip must be an str, not {0}.".format(
                type(level)
            ) + " Options are {0}".format(COMPRESSION_LEVELS["szip"])
            logger.error(msg)
            raise TypeError(msg)
    if not level in COMPRESSION_LEVELS[compression]:
        msg = (
            f"compression level {level} not supported for {compression}."
            + " Options are {0}".format(COMPRESSION_LEVELS[compression])
        )
        logger.error(msg)
        raise ValueError(msg)
    return compression, level


def recursive_hdf5_tree(group, lines=[]):
    if isinstance(group, (h5py._hl.group.Group, h5py._hl.files.File)):
        for key, value in group.items():
            lines.append("-{0}: {1}".format(key, value))
            recursive_hdf5_tree(value, lines)
    elif isinstance(group, h5py._hl.dataset.Dataset):
        for key, value in group.attrs.items():
            lines.append("\t-{0}: {1}".format(key, value))
    return "\n".join(lines)


def close_open_files():
    for obj in gc.get_objects():
        try:
            if isinstance(obj, h5py.File):
                msg = "Found HDF5 File object "
                logger.debug(msg)
                try:
                    msg = f"{obj.filename}, "
                    obj.flush()
                    obj.close()
                    msg += "Closed File"
                    logger.info(msg)
                except:
                    msg += f"{obj.filename} file already closed."
                    logger.info(msg)
        except:
            logger.debug(f"Object {type(obj)} does not have __class__")


def get_tree(parent):
    """
    Simple function to recursively print the contents of an hdf5 group
    Parameters
    ----------
    parent : :class:`h5py.Group`
        HDF5 (sub-)tree to print

    """
    lines = ["{0}:".format(parent.name), "=" * 20]
    if not isinstance(parent, (h5py.File, h5py.Group)):
        raise TypeError(
            "Provided object is not a h5py.File or h5py.Group " "object"
        )

    def fancy_print(name, obj):
        # lines.append(name)
        spacing = " " * 4 * (name.count("/") + 1)
        group_name = name[name.rfind("/") + 1 :]

        if isinstance(obj, h5py.Group):
            lines.append("{0}|- Group: {1}".format(spacing, group_name))
            lines.append("{0}{1}".format(spacing, (len(group_name) + 10) * "-"))
        elif isinstance(obj, h5py.Dataset):
            lines.append("{0}--> Dataset: {1}".format(spacing, group_name))
            lines.append("{0}{1}".format(spacing, (len(group_name) + 15) * "."))

    # lines.append(parent.name)
    parent.visititems(fancy_print)
    return "\n".join(lines)


def to_numpy_type(value):
    """
    Need to make the attributes friendly with Numpy and HDF5.

    For numbers and bool this is straight forward they are automatically
    mapped in h5py to a numpy type.

    But for strings this can be a challenge, especially a list of strings.

    HDF5 should only deal with ASCII characters or Unicode.  No binary data
    is allowed.
    """

    if value is None:
        return "none"
    # For now turn references into a generic string
    if isinstance(value, h5py.h5r.Reference):
        value = str(value)
    if isinstance(
        value,
        (
            str,
            np.str_,
            int,
            float,
            bool,
            complex,
            np.int_,
            np.float_,
            np.bool_,
            np.complex_,
        ),
    ):
        return value
    if isinstance(value, Iterable):
        if np.any([type(x) in [str, bytes, np.str_] for x in value]):
            return np.array(value, dtype="S")
        else:
            return np.array(value)
    else:
        raise TypeError("Type {0} not understood".format(type(value)))


def validate_name(name):
    """
    make sure the name has no spaces or slashes

    :param name: DESCRIPTION
    :type name: TYPE
    :return: DESCRIPTION
    :rtype: TYPE

    """

    return name.replace(" ", "_").replace("/", "_")


def from_numpy_type(value):
    """
    Need to make the attributes friendly with Numpy and HDF5.

    For numbers and bool this is straight forward they are automatically
    mapped in h5py to a numpy type.

    But for strings this can be a challenge, especially a list of strings.

    HDF5 should only deal with ASCII characters or Unicode.  No binary data
    is allowed.
    """

    if value is None:
        return "none"
    # For now turn references into a generic string
    if isinstance(value, h5py.h5r.Reference):
        value = str(value)
    if isinstance(
        value,
        (
            str,
            np.str_,
            int,
            float,
            bool,
            complex,
            np.int_,
            np.float_,
            np.bool_,
            np.complex_,
        ),
    ):
        return value
    if isinstance(value, Iterable):
        if np.any([type(x) in [bytes, np.bytes_] for x in value]):
            return np.array(value, dtype="U").tolist()
        else:
            return np.array(value).tolist()
    else:
        raise TypeError("Type {0} not understood".format(type(value)))


# =============================================================================
#
# =============================================================================
def inherit_doc_string(cls):
    for base in inspect.getmro(cls):
        if base.__doc__ is not None:
            cls.__doc__ = base.__doc__
            break
    return cls


def validate_name(name, pattern=None):
    """
    Validate name

    :param name: DESCRIPTION
    :type name: TYPE
    :param pattern: DESCRIPTION, defaults to None
    :type pattern: TYPE, optional
    :return: DESCRIPTION
    :rtype: TYPE

    """
    if name is None:
        return "unknown"
    return name.replace(" ", "_").replace(",", "")



# -*- coding: utf-8 -*-
"""
Exceptions raised by MTH5


Created on Wed May 13 19:07:21 2020

@author: jpeacock
"""

# Schema Error
class MTSchemaError(Exception):
    pass


class MTTimeError(Exception):
    pass


class MTH5Error(Exception):
    pass


class MTH5TableError(Exception):
    pass


class MTTSError(Exception):
    pass



# -*- coding: utf-8 -*-
"""
Module to read and parse native Phoenix Geophysics data formats of the 
MTU-5C Family

This module implements Streamed readers for segmented-decimated  time series
 formats of the MTU-5C family.
 
:author: Jorge Torres-Solis

Revised 2022 by J. Peacock 
"""


# =============================================================================
# Imports
# =============================================================================
import numpy as np

from struct import unpack_from

from mt_metadata.utils.mttime import MTime
from mth5.io.phoenix.readers import TSReaderBase
from mth5.timeseries import ChannelTS

# =============================================================================
class SubHeader:
    """
    Class for subheader of segmented files
    """

    def __init__(self, **kwargs):
        self.header_length = 32
        self._header = None

        for key, value in kwargs.items():
            setattr(self, key, value)
        self._unpack_dict = {
            "gps_time_stamp": {"dtype": "I", "index": 0},
            "n_samples": {"dtype": "I", "index": 4},
            "saturation_count": {"dtype": "H", "index": 8},
            "missing_count": {"dtype": "H", "index": 10},
            "value_min": {"dtype": "f", "index": 12},
            "value_max": {"dtype": "f", "index": 16},
            "value_mean": {"dtype": "f", "index": 20},
        }

    def __str__(self):
        lines = ["subheader information:", "-" * 30]
        for key in [
            "gps_time_stamp",
            "n_samples",
            "saturation_count",
            "missing_count",
            "value_min",
            "value_max",
            "value_mean",
        ]:
            lines.append(f"\t{key:<25}: {getattr(self, key)}")
        return "\n".join(lines)

    def __repr__(self):
        return self.__str__()

    def _has_header(self):
        if self._header is not None:
            return True
        return False

    def _unpack_value(self, key):
        if self._has_header():
            return unpack_from(
                self._unpack_dict[key]["dtype"],
                self._header,
                self._unpack_dict[key]["index"],
            )

    @property
    def gps_time_stamp(self):
        if self._has_header():
            return MTime(self._unpack_value("gps_time_stamp")[0])

    @property
    def n_samples(self):
        if self._has_header():
            return self._unpack_value("n_samples")[0]

    @property
    def saturation_count(self):
        if self._has_header():
            return self._unpack_value("saturation_count")[0]

    @property
    def missing_count(self):
        if self._has_header():
            return self._unpack_value("missing_count")[0]

    @property
    def value_min(self):
        if self._has_header():
            return self._unpack_value("value_min")[0]

    @property
    def value_max(self):
        if self._has_header():
            return self._unpack_value("value_max")[0]

    @property
    def value_mean(self):
        if self._has_header():
            return self._unpack_value("value_mean")[0]

    def unpack_header(self, stream):
        if self.header_length > 0:
            # be sure to read from the beginning of the file
            self._header = stream.read(self.header_length)
        else:
            return


class Segment(SubHeader):
    """
    A segment class to hold a single segment
    """

    def __init__(self, stream, **kwargs):

        super().__init__(**kwargs)
        self.stream = stream
        self.data = None

    def read_segment(self, metadata_only=False):
        """
        Read the whole file in

        :return: DESCRIPTION
        :rtype: TYPE

        """

        self.unpack_header(self.stream)
        if not metadata_only:
            self.data = np.fromfile(
                self.stream, dtype=np.float32, count=self.n_samples
            )

    @property
    def segment_start_time(self):
        """
        estimate the segment start time based on sequence number

        :return: DESCRIPTION
        :rtype: TYPE

        """

        return self.gps_time_stamp

    @property
    def segment_end_time(self):
        """
        estimate end time

        :return: DESCRIPTION
        :rtype: TYPE

        """

        return self.segment_start_time + ((self.n_samples) / self.sample_rate)


class DecimatedSegmentedReader(TSReaderBase):
    """
    Class to create a streamer for segmented decimated time series,
    i.e. 'td_24k'

    These files have a sub header

    """

    def __init__(self, path, num_files=1, report_hw_sat=False, **kwargs):
        # Init the base class
        super().__init__(
            path,
            num_files=num_files,
            header_length=128,
            report_hw_sat=report_hw_sat,
            **kwargs,
        )

        self.unpack_header(self.stream)
        self.sub_header = SubHeader()
        self.subheader = {}

    def read_segment(self, metadata_only=False):
        """
        Read in a single segment

        :param metadata_only: DESCRIPTION, defaults to False
        :type metadata_only: TYPE, optional
        :return: DESCRIPTION
        :rtype: TYPE

        """

        kwargs = {
            "instrument_type": self.instrument_type,
            "instrument_serial_number": self.instrument_serial_number,
            "latitude": self.gps_lat,
            "longitude": self.gps_long,
            "elevation": self.gps_elevation,
            "sample_rate": self.sample_rate,
            "channel_id": self.channel_id,
            "channel_type": self.channel_type,
            "segment": 0,
        }

        segment = Segment(self.stream, **kwargs)
        segment.read_segment(metadata_only=metadata_only)

        return segment

    def to_channel_ts(self):
        """
        convert to a ChannelTS object

        :return: DESCRIPTION
        :rtype: TYPE

        """

        segment = self.read_segment()
        ch_metadata = self.channel_metadata()
        ch_metadata.time_period.start = segment.segment_start_time.isoformat()

        return ChannelTS(
            channel_type=ch_metadata.type,
            data=segment.data,
            channel_metadata=ch_metadata,
            run_metadata=self.run_metadata(),
            station_metadata=self.station_metadata(),
        )


class DecimatedSegmentCollection(TSReaderBase):
    """
    Class to create a streamer for segmented decimated time series,
    i.e. 'td_24k'

    These files have a sub header

    """

    def __init__(self, path, num_files=1, report_hw_sat=False, **kwargs):
        # Init the base class
        super().__init__(
            path,
            num_files=num_files,
            header_length=128,
            report_hw_sat=report_hw_sat,
            **kwargs,
        )

        self.unpack_header(self.stream)
        self.sub_header = SubHeader()
        self.subheader = {}

    def read_segments(self, metadata_only=False):
        """
        Read the whole file in

        :return: DESCRIPTION
        :rtype: TYPE

        """

        kwargs = {
            "instrument_type": self.instrument_type,
            "instrument_serial_number": self.instrument_serial_number,
            "latitude": self.gps_lat,
            "longitude": self.gps_long,
            "elevation": self.gps_elevation,
            "sample_rate": self.sample_rate,
            "channel_id": self.channel_id,
            "channel_type": self.channel_type,
            "segment": 0,
        }
        segments = []
        count = 1
        while True:
            try:
                kwargs["segment"] = count
                segment = Segment(self.stream, **kwargs)
                segment.read_segment(metadata_only=metadata_only)
                segments.append(segment)

                count += 1
            except:
                break
        self.logger.info(f"Read {count - 1} segments")

        return segments

    def to_channel_ts(self):
        """
        convert to a ChannelTS object

        :return: DESCRIPTION
        :rtype: TYPE

        """

        seq_list = []
        for seq in self.read_segments():
            ch_metadata = self.channel_metadata()
            ch_metadata.time_period.start = seq.gps_time_stamp.isoformat()

            seq_list.append(
                ChannelTS(
                    channel_type=ch_metadata.type,
                    data=seq.data,
                    channel_metadata=ch_metadata,
                    run_metadata=self.run_metadata(),
                    station_metadata=self.station_metadata(),
                )
            )
        return seq_list



# -*- coding: utf-8 -*-
"""
Created on Thu Aug 27 16:54:09 2020

:author: Jared Peacock

:license: MIT

"""

# =============================================================================
# Imports
# =============================================================================
from pathlib import Path
from urllib.request import Request, urlopen
from urllib.error import HTTPError
import xml.etree.ElementTree as ET
from collections import OrderedDict

import numpy as np

from mth5.utils.mth5_logger import setup_logger
from mt_metadata.timeseries import Magnetic, Electric, Run, Station, Survey

# =============================================================================
#  Metadata for usgs ascii file
# =============================================================================


class AsciiMetadata:
    """
    Container for all the important metadata in a USGS ascii file.

    ========================= =================================================
    Attributes                Description
    ========================= =================================================
    survey_id                  Survey name
    site_id                    Site name
    run_id                     Run number
    site_latitude              Site latitude in decimal degrees WGS84
    site_longitude             Site longitude in decimal degrees WGS84
    site_elevation             Site elevation according to national map meters
    start                      Start time of station YYYY-MM-DDThh:mm:ss UTC
    end                        Stop time of station YYYY-MM-DDThh:mm:ss UTC
    sample_rate                Sampling rate samples/second
    n_samples                  Number of samples
    n_channels                 Number of channels
    coordinate_system          [ Geographic North | Geomagnetic North ]
    chn_settings               Channel settings, see below
    missing_data_flag          Missing data value
    ========================= =================================================

    :chn_settings:

    ========================= =================================================
    Keys                      Description
    ========================= =================================================
    ChnNum                    site_id+channel number
    ChnID                     Component [ ex | ey | hx | hy | hz ]
    InstrumentID              Data logger + sensor number
    Azimuth                   Setup angle of componet in degrees relative to
                              coordinate_system
    Dipole_Length             Dipole length in meters
    ========================= =================================================


    """

    def __init__(self, fn=None, **kwargs):

        self.logger = setup_logger(f"{__name__}.{self.__class__.__name__}")

        self.fn = fn
        self.missing_data_flag = np.NaN
        self.coordinate_system = None
        self._metadata_len = 30
        self._survey_metadata = Survey()
        self._station_metadata = Station()
        self._run_metadata = Run()
        self.ex_metadata = Electric(component="ex")
        self.ey_metadata = Electric(component="ey")
        self.hx_metadata = Magnetic(component="hx")
        self.hy_metadata = Magnetic(component="hy")
        self.hz_metadata = Magnetic(component="hz")

        self.channel_order = ["hx", "ex", "hy", "ey", "hz"]

        self._key_dict = OrderedDict(
            **{
                "SurveyID": "survey_id",
                "SiteID": "site_id",
                "RunID": "run_id",
                "SiteLatitude": "latitude",
                "SiteLongitude": "longitude",
                "SiteElevation": "elevation",
                "AcqStartTime": "start",
                "AcqStopTime": "end",
                "AcqSmpFreq": "sample_rate",
                "AcqNumSmp": "n_samples",
                "Nchan": "n_channels",
                "Channel coordinates relative to geographic north": "",
                "ChnSettings": "",
                "MissingDataFlag": "missing_data_flag",
                "DataSet": "data_set",
            }
        )

        self._chn_dict = OrderedDict(
            **{
                "ChnNum": "channel_number",
                "ChnID": "component",
                "InstrumentID": "sensor.id",
                "Azimuth": "measurement_azimuth",
                "Dipole_Length": "dipole_length",
            }
        )
        self._chn_fmt = {
            "ChnNum": "<8",
            "ChnID": "<6",
            "InstrumentID": "<12",
            "Azimuth": ">7.1f",
            "Dipole_Length": ">14.1f",
        }

        for key in kwargs.keys():
            setattr(self, key, kwargs[key])

    def __str__(self):
        return "\n".join(self.write_metadata())

    def __repr__(self):
        return self.__str__()

    @property
    def fn(self):
        return self._fn

    @fn.setter
    def fn(self, value):
        if value is not None:
            self._fn = Path(value)
        else:
            self._fn = None

    @property
    def file_size(self):
        if self.fn is not None:
            if self.fn.exists():
                return self.fn.stat().st_size

    @property
    def survey_id(self):
        return self._survey_metadata.id

    @survey_id.setter
    def survey_id(self, value):
        self._survey_metadata.id = value

    @property
    def run_id(self):
        return self._run_metadata.id

    @run_id.setter
    def run_id(self, value):
        self._run_metadata.id = value

    @property
    def site_id(self):
        return self._station_metadata.id

    @site_id.setter
    def site_id(self, station):
        self._station_metadata.id = station

    @property
    def latitude(self):
        return self._station_metadata.location.latitude

    @latitude.setter
    def latitude(self, lat):
        self._station_metadata.location.latitude = lat

    @property
    def longitude(self):
        return self._station_metadata.location.longitude

    @longitude.setter
    def longitude(self, lon):
        self._station_metadata.location.longitude = lon

    @property
    def elevation(self):
        """
        get elevation from national map
        """
        # the url for national map elevation query
        nm_url = Request(
            r"https://nationalmap.gov/epqs/pqs.php?x={0:.5f}&y={1:.5f}&units=Meters&output=xml".format(
                self.longitude, self.latitude
            )
        )

        # call the url and get the response
        try:
            response = urlopen(nm_url)
        except HTTPError:
            self.logger.error(
                "could not connect to get elevation from national map."
            )
            self.logger.debug(nm_url.format(self.longitude, self.latitude))
            return self.station_metadata.location.elevation

        # read the xml response and convert to a float
        info = ET.ElementTree(ET.fromstring(response.read()))
        info = info.getroot()
        for elev in info.iter("Elevation"):
            nm_elev = float(elev.text)
        return nm_elev

    @elevation.setter
    def elevation(self, value):
        self.station_metadata.location.elevation = value

    @property
    def start(self):
        return self._station_metadata.time_period.start

    @start.setter
    def start(self, time_string):
        self.station_metadata.time_period.start = time_string

    @property
    def end(self):
        return self._station_metadata.time_period.end

    @end.setter
    def end(self, time_string):
        self._station_metadata.time_period.end = time_string

    @property
    def n_channels(self):
        return self._chn_num

    @n_channels.setter
    def n_channels(self, n_channel):
        try:
            self._chn_num = int(n_channel)
        except ValueError:
            self.logger.warning(
                f"{n_channel} is not a number, setting n_channels to 0"
            )

    @property
    def sample_rate(self):
        return self._run_metadata.sample_rate

    @sample_rate.setter
    def sample_rate(self, df):
        self._run_metadata.sample_rate = float(df)

    @property
    def n_samples(self):
        return self._n_samples

    @n_samples.setter
    def n_samples(self, n_samples):
        self._n_samples = int(n_samples)

    @property
    def survey_metadata(self):
        return self._survey_metadata

    @survey_metadata.setter
    def survey_metadata(self, value):
        if isinstance(value, Survey):
            self._survey_metadata.update(value)
        else:
            raise TypeError(
                "Input must be a mt_metadata.timeseries.Survey instance"
            )

    @property
    def station_metadata(self):
        return self._station_metadata

    @station_metadata.setter
    def station_metadata(self, value):
        if isinstance(value, Station):
            self._station_metadata.update(value)
        else:
            raise TypeError(
                "Input must be a mt_metadata.timeseries.Station instance"
            )

    @property
    def run_metadata(self):
        return self._run_metadata

    @run_metadata.setter
    def run_metadata(self, value):
        if isinstance(value, Run):
            self._run_metadata.update(value)
        else:
            raise TypeError(
                "Input must be a mt_metadata.timeseries.Run instance"
            )

    def get_component_info(self, comp):
        """

        :param comp: DESCRIPTION
        :type comp: TYPE
        :return: DESCRIPTION
        :rtype: TYPE

        """

        for key, kdict in self.channel_dict.items():
            if kdict["ChnID"].lower() == comp.lower():
                return kdict

        return None

    def read_metadata(self, fn=None, meta_lines=None):
        """
        Read in a meta from the raw string or file.  Populate all metadata
        as attributes.

        :param fn: full path to USGS ascii file
        :type fn: string

        :param meta_lines: lines of metadata to read
        :type meta_lines: list
        """
        chn_find = False

        if fn is not None:
            self.fn = fn
        if self.fn is not None:
            with self.fn.open("r") as fid:
                meta_lines = [
                    fid.readline() for ii in range(self._metadata_len)
                ]
        for ii, line in enumerate(meta_lines):
            if "DataSet" in line:
                break

            if line.find(":") > 0:
                key, value = line.strip().split(":", 1)
                value = value.strip()

                if len(value) < 1:
                    chn_find = True
                    continue

                attr = self._key_dict[key]
                setattr(self, attr, value)

            elif "coordinate" in line:
                self.coordinate_system = " ".join(line.strip().split()[-2:])
            else:
                if chn_find is True:
                    if "chnnum" in line.lower():
                        ch_keys = dict(
                            [
                                (kk, ii)
                                for ii, kk in enumerate(line.strip().split())
                            ]
                        )
                    else:
                        line_list = line.strip().split()
                        if len(line_list) == len(ch_keys.keys()):
                            ch = getattr(
                                self,
                                f"{line_list[ch_keys['ChnID']].lower()}_metadata",
                            )

                            ch.channel_number = line_list[ch_keys["ChnNum"]]
                            ch.measurement_azimuth = line_list[
                                ch_keys["Azimuth"]
                            ]
                            if ch.type in ["electric"]:
                                ch.dipole_length = line_list[
                                    ch_keys["Dipole_Length"]
                                ]
                            else:
                                ch.sensor.id = line_list[
                                    ch_keys["InstrumentID"]
                                ]
                            self.run_metadata.data_logger.id = line_list[
                                ch_keys["InstrumentID"]
                            ]

                            ch.time_period.start = self.start
                            ch.time_period.end = self.end
                            ch.sample_rate = self.sample_rate
                            self._run_metadata.add_channel(ch)

                        else:
                            self.logger.warning("Not sure what line this is")

        self._run_metadata.time_period.start = self.start
        self._run_metadata.time_period.end = self.end
        self._run_metadata.sample_rate = self.sample_rate

        self._station_metadata.add_run(self.run_metadata)

        return ii + 1

    def write_metadata(self, chn_list=["Ex", "Ey", "Hx", "Hy", "Hz"]):
        """
        Write out metadata in the format of USGS ascii.

        :return: list of metadate lines.

        .. note:: meant to use '\n'.join(lines) to write out in a file.

        """

        lines = []
        for key, attr in self._key_dict.items():
            if key in ["ChnSettings"]:
                lines.append("{0}:".format(key))
                lines.append(" ".join(self._chn_dict.keys()))
                for chn_key in self.channel_order:
                    chn_line = []
                    ch = getattr(self, f"{chn_key}_metadata")
                    for comp_key, comp_attr in self._chn_dict.items():
                        try:
                            value = ch.get_attr_from_name(comp_attr)
                            if value is None:
                                value = self.run_metadata.data_logger.id
                        except AttributeError:
                            value = 0
                        chn_line.append(f"{value:{self._chn_fmt[comp_key]}}")
                    lines.append("".join(chn_line))

            elif key in ["DataSet"]:
                lines.append("{0}:".format(key))
                return lines
            else:
                try:
                    value = getattr(self, attr)
                    lines.append(f"{key}: {value}")
                except AttributeError:
                    lines.append(f"{key}")

        return lines



# -*- coding: utf-8 -*-
"""
Created on Tue Jun  8 17:58:47 2021

:copyright: 
    Jared Peacock (jpeacock@usgs.gov)

:license: MIT

"""

import unittest
from pathlib import Path
import numpy as np
import numpy.testing as npt
from mth5 import mth5

from mt_metadata.timeseries.stationxml import XMLInventoryMTExperiment
from mt_metadata import STATIONXML_FAP

fn_path = Path(__file__).parent


class TestFAPMTH5(unittest.TestCase):
    """
    Test making an MTH5 file from a FAP filtered StationXML

    """

    @classmethod
    def setUpClass(self):
        self.translator = XMLInventoryMTExperiment()
        self.experiment = self.translator.xml_to_mt(
            stationxml_fn=STATIONXML_FAP
        )

        self.fn = fn_path.joinpath("from_fap_stationxml.h5")
        # if self.fn.exists():
        #     self.fn.unlink()

        self.m = mth5.MTH5(file_version="0.1.0")
        self.m.open_mth5(self.fn, mode="a")
        self.m.from_experiment(self.experiment, 0)

    def test_has_survey(self):
        self.assertEqual(self.m.has_group("Survey"), True)

    def test_has_station(self):
        self.assertEqual(self.m.has_group("Survey/Stations"), True)
        self.assertEqual(self.m.has_group("Survey/Stations/FL001"), True)

    def test_has_run_a(self):
        self.assertEqual(self.m.has_group("Survey/Stations/FL001/a"), True)

    def test_has_run_b(self):
        self.assertEqual(self.m.has_group("Survey/Stations/FL001/b"), True)

    def test_has_hx_a(self):
        self.assertEqual(self.m.has_group("Survey/Stations/FL001/a/hx"), True)

    def test_has_hx_b(self):
        self.assertEqual(self.m.has_group("Survey/Stations/FL001/b/hx"), True)

    def test_has_fap_table(self):

        self.assertEqual(
            self.m.has_group("Survey/Filters/fap/frequency response table_00"),
            True,
        )

    def test_has_coefficient_filter(self):
        self.assertEqual(
            self.m.has_group(
                "Survey/Filters/coefficient/v to counts (electric)"
            ),
            True,
        )

    def test_get_channel(self):
        self.hx = self.m.get_channel("FL001", "a", "hx")
        fnames = [f.name for f in self.hx.channel_response_filter.filters_list]

        self.assertIn("frequency response table_00", fnames)
        self.assertIn("v to counts (electric)", fnames)

    def test_fap(self):
        self.hx = self.m.get_channel("FL001", "a", "hx")
        fap = self.hx.channel_response_filter.filters_list[0]
        fap_exp = self.experiment.surveys[0].filters[
            "frequency response table_00"
        ]

        self.assertTrue(np.allclose(fap.frequencies, fap_exp.frequencies, 7))
        self.assertTrue(np.allclose(fap.amplitudes, fap_exp.amplitudes, 7))
        self.assertTrue(np.allclose(fap.phases, np.deg2rad(fap_exp.phases), 7))

        npt.assert_almost_equal(fap.frequencies, fap_exp.frequencies, 7)
        npt.assert_almost_equal(fap.amplitudes, fap_exp.amplitudes, 7)
        npt.assert_almost_equal(fap.phases, np.deg2rad(fap_exp.phases), 7)

        for k in ["gain", "units_in", "units_out", "name", "comments"]:
            self.assertEqual(getattr(fap, k), getattr(fap_exp, k))

    def test_coefficient(self):
        self.hx = self.m.get_channel("FL001", "a", "hx")
        coeff = self.hx.channel_response_filter.filters_list[1]
        coeff_exp = self.experiment.surveys[0].filters[
            "v to counts (electric)"
        ]

        self.assertDictEqual(
            coeff.to_dict(single=True), coeff_exp.to_dict(single=True)
        )

    @classmethod
    def tearDownClass(self):
        self.m.close_mth5()
        self.fn.unlink()



# -*- coding: utf-8 -*-
"""
.. module:: timeseries
   :synopsis: Deal with MT time series

.. todo:: Check the conversion to netcdf.  There are some weird serializations of
lists and arrays that goes on, seems easiest to convert all lists to strings and then
convert them back if read in.


:copyright:
    Jared Peacock (jpeacock@usgs.gov)

:license:
    MIT
"""

# ==============================================================================
# Imports
# ==============================================================================
import inspect

import xarray as xr
import numpy as np

from matplotlib import pyplot as plt

from mt_metadata import timeseries as metadata
from mt_metadata.utils.mttime import MTime
from mt_metadata.utils.list_dict import ListDict
from mt_metadata.timeseries.filters import ChannelResponseFilter

from .channel_ts import ChannelTS
from mth5.utils.mth5_logger import setup_logger

from obspy.core import Stream

# =============================================================================
# make a dictionary of available metadata classes
# =============================================================================
meta_classes = dict(inspect.getmembers(metadata, inspect.isclass))


# =============================================================================
# run container
# =============================================================================
class RunTS:
    """
    holds all run ts in one aligned array

    components --> {'ex': ex_xarray, 'ey': ey_xarray}

    ToDo, have a single Survey object under the hood and properties to other
    metadata objects for get/set.

    """

    def __init__(
        self,
        array_list=None,
        run_metadata=None,
        station_metadata=None,
        survey_metadata=None,
    ):

        self.logger = setup_logger(f"{__name__}.{self.__class__.__name__}")
        self._survey_metadata = self._initialize_metadata()
        self._dataset = xr.Dataset()
        self._filters = {}

        self.survey_metadata = survey_metadata
        self.station_metadata = station_metadata
        self.run_metadata = run_metadata

        # load the arrays first this will write run and station metadata
        if array_list is not None:
            self.dataset = array_list

    def __str__(self):
        s_list = [
            f"Survey:      {self.survey_metadata.id}",
            f"Station:     {self.station_metadata.id}",
            f"Run:         {self.run_metadata.id}",
            f"Start:       {self.start}",
            f"End:         {self.end}",
            f"Sample Rate: {self.sample_rate}",
            f"Components:  {self.channels}",
        ]
        return "\n\t".join(["RunTS Summary:"] + s_list)

    def __repr__(self):
        return self.__str__()

    def _initialize_metadata(self):
        """
        Create a single `Survey` object to store all metadata

        :param channel_type: DESCRIPTION
        :type channel_type: TYPE
        :return: DESCRIPTION
        :rtype: TYPE

        """

        survey_metadata = metadata.Survey(id="0")
        survey_metadata.stations.append(metadata.Station(id="0"))
        survey_metadata.stations[0].runs.append(metadata.Run(id="0"))

        return survey_metadata

    def _validate_run_metadata(self, run_metadata):
        """
        validate run metadata

        """

        if not isinstance(run_metadata, metadata.Run):
            if isinstance(run_metadata, dict):
                if "run" not in [cc.lower() for cc in run_metadata.keys()]:
                    run_metadata = {"Run": run_metadata}
                r_metadata = metadata.Run()
                r_metadata.from_dict(run_metadata)
                self.logger.debug("Loading from metadata dict")
                return r_metadata
            else:
                msg = "input metadata must be type %s or dict, not %s"
                self.logger.error(
                    msg, type(self.run_metadata), type(run_metadata)
                )
                raise TypeError(
                    msg % (type(self.run_metadata), type(run_metadata))
                )
        return run_metadata.copy()

    def _validate_station_metadata(self, station_metadata):
        """
        validate station metadata
        """

        if not isinstance(station_metadata, metadata.Station):
            if isinstance(station_metadata, dict):
                if "station" not in [
                    cc.lower() for cc in station_metadata.keys()
                ]:
                    station_metadata = {"Station": station_metadata}

                st_metadata = metadata.Station()
                st_metadata.from_dict(station_metadata)
                self.logger.debug("Loading from metadata dict")
                return st_metadata
            else:
                msg = "input metadata must be type {0} or dict, not {1}".format(
                    type(self.station_metadata), type(station_metadata)
                )
                self.logger.error(msg)
                raise TypeError(msg)

        return station_metadata.copy()

    def _validate_survey_metadata(self, survey_metadata):
        """
        validate station metadata
        """

        if not isinstance(survey_metadata, metadata.Survey):
            if isinstance(survey_metadata, dict):
                if "station" not in [
                    cc.lower() for cc in survey_metadata.keys()
                ]:
                    survey_metadata = {"Survey": survey_metadata}

                sv_metadata = metadata.Station()
                sv_metadata.from_dict(survey_metadata)
                self.logger.debug("Loading from metadata dict")
                return sv_metadata
            else:
                msg = "input metadata must be type {0} or dict, not {1}".format(
                    type(self.survey_metadata), type(survey_metadata)
                )
                self.logger.error(msg)
                raise TypeError(msg)

        return survey_metadata.copy()

    def _validate_array_list(self, array_list):
        """check to make sure all entries are a :class:`ChannelTS` object"""

        if not isinstance(array_list, (tuple, list)):
            msg = f"array_list must be a list or tuple, not {type(array_list)}"
            self.logger.error(msg)
            raise TypeError(msg)

        valid_list = []
        station_metadata = metadata.Station()
        run_metadata = metadata.Run()
        channels = ListDict()

        for index, item in enumerate(array_list):
            if not isinstance(item, (ChannelTS, xr.DataArray)):
                msg = f"array entry {index} must be ChannelTS object not {type(item)}"
                self.logger.error(msg)
                raise TypeError(msg)
            if isinstance(item, ChannelTS):
                valid_list.append(item.to_xarray())

                # if a channelTS is input then it comes with run and station metadata
                # use those first, then the user can update later.

                if item.station_metadata.id not in ["0", None]:
                    if station_metadata.id not in ["0", None]:
                        station_metadata.update(
                            item.station_metadata, match=["id"]
                        )
                    else:
                        station_metadata.update(item.station_metadata)

                if item.run_metadata.id not in ["0", None]:
                    if run_metadata.id not in ["0", None]:
                        run_metadata.update(item.run_metadata, match=["id"])
                    else:
                        run_metadata.update(item.run_metadata)

                channels.append(item.channel_metadata)

                # get the filters from the channel
                if item.channel_response_filter.filters_list != []:
                    for ff in item.channel_response_filter.filters_list:
                        self._filters[ff.name] = ff

            else:
                valid_list.append(item)

        # need to make sure that the station metadata was actually updated,
        # should have an ID.
        run_metadata.channels = channels
        if station_metadata.id not in ["0", None]:
            station_metadata.runs = ListDict()
            station_metadata.runs.append(run_metadata)
            # need to add the other runs that are in the metadata for
            # completeness.
            for run in self.station_metadata.runs:
                if run.id not in [run_metadata.id, "0", None]:
                    station_metadata.runs.append(run)
            self.station_metadata = station_metadata
        # if the run metadata was updated
        elif run_metadata.id not in ["0", None]:
            self.run_metadata = run_metadata
        # if the run metadata or station metadata was not updated from channel
        # metadata, then update just the channels.
        else:
            self.run_metadata.channels = channels

        # probably should test for sampling rate.
        sr_test = dict(
            [(item.component, (item.sample_rate)) for item in valid_list]
        )

        if len(set([v for k, v in sr_test.items()])) != 1:
            msg = f"sample rates are not all the same {sr_test}"
            self.logger.error(msg)
            raise ValueError(msg)

        return valid_list

    def _get_channel_response_filter(self, ch_name):
        """
        Get the channel response filter from the filter dictionary

        :param ch_name: DESCRIPTION
        :type ch_name: TYPE
        :return: DESCRIPTION
        :rtype: TYPE

        """

        filter_list = []
        if ch_name in self.dataset.keys():

            for filter_name in self.dataset[ch_name].attrs["filter.name"]:
                try:
                    filter_list.append(self.filters[filter_name])
                except KeyError:
                    self.logger.debug(
                        f"Could not find {filter_name} in filters"
                    )

        return ChannelResponseFilter(filters_list=filter_list)

    def __getattr__(self, name):
        # change to look for keys directly and use type to set channel type
        if name in self.dataset.keys():

            ch_response_filter = self._get_channel_response_filter(name)
            # if cannot get filters, but the filters name indicates that
            # filters should be there don't input the channel response filter
            # cause then an empty filters_list will set filter.name to []
            if ch_response_filter.filters_list == []:
                ch_response_filter = None

            return ChannelTS(
                self.dataset[name].attrs["type"],
                self.dataset[name],
                run_metadata=self.run_metadata.copy(),
                station_metadata=self.station_metadata.copy(),
                channel_response_filter=ch_response_filter,
            )
        else:
            # this is a hack for now until figure out who is calling shape, size
            if name[0] == "_":
                return None
            if name not in ["shape", "size"]:
                try:
                    return super().__getattribute__(name)
                except AttributeError:
                    msg = f"RunTS has no attribute {name}"
                    self.logger.error(msg)
                    raise NameError(msg)

    ### Properties ------------------------------------------------------------
    @property
    def survey_metadata(self):
        """
        survey metadata
        """
        return self._survey_metadata

    @survey_metadata.setter
    def survey_metadata(self, survey_metadata):
        """

        :param survey_metadata: survey metadata object or dictionary
        :type survey_metadata: :class:`mt_metadata.timeseries.Survey` or dict

        """

        if survey_metadata is not None:
            survey_metadata = self._validate_survey_metadata(survey_metadata)
            self._survey_metadata.update(
                self._validate_survey_metadata(survey_metadata)
            )

    @property
    def station_metadata(self):
        """
        station metadata
        """

        return self.survey_metadata.stations[0]

    @station_metadata.setter
    def station_metadata(self, station_metadata):
        """
        set station metadata from a valid input
        """

        if station_metadata is not None:
            station_metadata = self._validate_station_metadata(station_metadata)

            runs = ListDict()
            if self.run_metadata.id not in ["0", 0]:
                runs.append(self.run_metadata.copy())
            runs.extend(station_metadata.runs)
            if len(runs) == 0:
                runs[0] = metadata.Run(id="0")

            # be sure there is a level below
            if len(runs[0].channels) == 0:
                ch_metadata = metadata.Auxiliary()
                ch_metadata.type = "auxiliary"
                runs[0].channels.append(ch_metadata)

            stations = ListDict()
            stations.append(station_metadata)
            stations[0].runs = runs

            self.survey_metadata.stations = stations

    @property
    def run_metadata(self):
        """
        station metadata
        """
        run_metadata = self.survey_metadata.stations[0].runs[0]

        return run_metadata

    @run_metadata.setter
    def run_metadata(self, run_metadata):
        """
        set run metadata from a valid input
        """

        if run_metadata is not None:
            run_metadata = self._validate_run_metadata(run_metadata)
            runs = ListDict()
            runs.append(run_metadata)
            runs.extend(
                self.station_metadata.runs, skip_keys=[run_metadata.id, "0"]
            )
            self._survey_metadata.stations[0].runs = runs

    def has_data(self):
        """check to see if there is data"""
        if len(self.channels) > 0:
            return True
        return False

    @property
    def summarize_metadata(self):
        """

        Get a summary of all the metadata

        :return: A summary of all channel metadata in one place
        :rtype: dictionary

        """
        meta_dict = {}
        for comp in self.dataset.data_vars:
            for mkey, mvalue in self.dataset[comp].attrs.items():
                meta_dict[f"{comp}.{mkey}"] = mvalue

        return meta_dict

    def validate_metadata(self):
        """
        Check to make sure that the metadata matches what is in the data set.

        updates metadata from the data.

        Check the start and end times, channels recorded

        """

        # check sampling rate
        if self.has_data():
            # check start time
            if self.start != self.run_metadata.time_period.start:
                if (
                    self.run_metadata.time_period.start
                    != "1980-01-01T00:00:00+00:00"
                ):
                    msg = (
                        f"start time of dataset {self.start} does not "
                        f"match metadata start {self.run_metadata.time_period.start} "
                        f"updating metatdata value to {self.start}"
                    )
                    self.logger.warning(msg)
                self.run_metadata.time_period.start = self.start.iso_str

            # check end time
            if self.end != self.run_metadata.time_period.end:
                if (
                    self.run_metadata.time_period.end
                    != "1980-01-01T00:00:00+00:00"
                ):
                    msg = (
                        f"end time of dataset {self.end} does not "
                        f"match metadata end {self.run_metadata.time_period.end} "
                        f"updating metatdata value to {self.end}"
                    )
                    self.logger.warning(msg)
                self.run_metadata.time_period.end = self.end.iso_str
            if self.sample_rate != self.run_metadata.sample_rate:
                if self.run_metadata.sample_rate == 0.0:
                    pass
                elif self.run_metadata.sample_rate is not None:
                    msg = (
                        f"sample rate of dataset {self.sample_rate} does not "
                        f"match metadata sample rate {self.run_metadata.sample_rate} "
                        f"updating metatdata value to {self.sample_rate}"
                    )
                    self.logger.warning(msg)
                self.run_metadata.sample_rate = self.sample_rate

            if self.run_metadata.id not in self.station_metadata.runs.keys():
                self.station_metadata.runs[0].update(self.run_metadata)

            self.station_metadata.update_time_period()
            self.survey_metadata.update_time_period()

    def set_dataset(self, array_list, align_type="outer"):
        """

        :param array_list: list of xarrays
        :type array_list: list of :class:`mth5.timeseries.ChannelTS` objects
        :param align_type: how the different times will be aligned
            * ’outer’: use the union of object indexes
            * ’inner’: use the intersection of object indexes
            * ’left’: use indexes from the first object with each dimension
            * ’right’: use indexes from the last object with each dimension
            * ’exact’: instead of aligning, raise ValueError when indexes to
            be aligned are not equal
            * ’override’: if indexes are of same size, rewrite indexes to
            be those of the first object with that dimension. Indexes for
            the same dimension must have the same size in all objects.
        :type align_type: string

        """
        if isinstance(array_list, (list, tuple)):
            x_array_list = self._validate_array_list(array_list)

            # first need to align the time series.
            x_array_list = xr.align(*x_array_list, join=align_type)

            # input as a dictionary
            xdict = dict([(x.component.lower(), x) for x in x_array_list])
            self._dataset = xr.Dataset(xdict)

        elif isinstance(array_list, xr.Dataset):
            self._dataset = array_list

        self.validate_metadata()
        self._dataset.attrs.update(self.run_metadata.to_dict(single=True))

    def add_channel(self, channel):
        """
        Add a channel to the dataset, can be an :class:`xarray.DataArray` or
        :class:`mth5.timeseries.ChannelTS` object.

        Need to be sure that the coordinates and dimensions are the same as the
        existing dataset, namely coordinates are time, and dimensions are the same,
        if the dimesions are larger than the existing dataset then the added channel
        will be clipped to the dimensions of the existing dataset.

        If the start time is not the same nan's will be placed at locations where the
        timing does not match the current start time.  This is a feature of xarray.


        :param channel: a channel xarray or ChannelTS to add to the run
        :type channel: :class:`xarray.DataArray` or :class:`mth5.timeseries.ChannelTS`


        """

        if isinstance(channel, xr.DataArray):
            c = ChannelTS()
            c.ts = channel
        elif isinstance(channel, ChannelTS):
            c = channel
            self.run_metadata.channels.append(c.channel_metadata)
            for ff in c.channel_response_filter.filters_list:
                self._filters[ff.name] = ff

        else:
            raise ValueError(
                "Input Channel must be type xarray.DataArray or ChannelTS"
            )

        ### need to validate the channel to make sure sample rate is the same
        if c.sample_rate != self.sample_rate:
            msg = (
                f"Channel sample rate is not correct, current {self.sample_rate} "
                + f"input {c.sample_rate}"
            )
            self.logger.error(msg)
            raise ValueError(msg)

        ### should probably check for other metadata like station and run?
        if len(self.dataset.dims) == 0:
            self.dataset = c._ts.to_dataset()
        else:
            self.dataset = xr.merge([self.dataset, c._ts.to_dataset()])

    @property
    def dataset(self):
        """:class:`xarray.Dataset`"""
        return self._dataset

    @dataset.setter
    def dataset(self, array_list):
        """Set the dataset"""
        msg = (
            "Data will be aligned using the min and max time. "
            "If that is not correct use set_dataset and change the alignment type."
        )
        self.logger.debug(msg)
        self.set_dataset(array_list)

    @property
    def start(self):
        """Start time UTC"""
        if self.has_data():
            return MTime(self.dataset.coords["time"].to_index()[0].isoformat())
        return self.run_metadata.time_period.start

    @property
    def end(self):
        """End time UTC"""
        if self.has_data():
            return MTime(self.dataset.coords["time"].to_index()[-1].isoformat())
        return self.run_metadata.time_period.end

    @property
    def sample_rate(self):
        """
        Sample rate, this is estimated by the mdeian difference between
        samples in time, if data is present. Otherwise return the metadata
        sample rate.
        """
        if self.has_data():
            try:
                return round(
                    1.0
                    / np.float64(
                        (
                            np.median(
                                np.diff(self.dataset.coords["time"].to_index())
                                / np.timedelta64(1, "s")
                            )
                        )
                    ),
                    0,
                )
            except AttributeError:
                self.logger.warning(
                    "Something weird happend with xarray time indexing"
                )

                raise ValueError(
                    "Something weird happend with xarray time indexing"
                )
        return self.run_metadata.sample_rate

    @property
    def sample_interval(self):
        """
        Sample interval = 1 / sample_rate
        :return: DESCRIPTION
        :rtype: TYPE

        """

        if self.sample_rate != 0:
            return 1.0 / self.sample_rate
        return 0.0

    @property
    def channels(self):
        """List of channel names in dataset"""
        return [cc for cc in list(self.dataset.data_vars)]

    @property
    def filters(self):
        """Dictionary of filters used by the channels"""
        return self._filters

    @filters.setter
    def filters(self, value):
        """
        a dictionary of filters found in the channel objects.

        Should use the dictionary methods to update a dictionary.

        :param value: dictionary of :module:`mt_metadata.timeseries.filters`
        objects
        :type value: dictionary
        :raises TypeError: If input is anything other than a dictionary

        """
        if not isinstance(value, dict):
            raise TypeError("input must be a dictionary")

        self._filters = value

    def to_obspy_stream(self):
        """
        convert time series to an :class:`obspy.core.Stream` which is like a
        list of :class:`obspy.core.Trace` objects.

        :return: An Obspy Stream object from the time series data
        :rtype: :class:`obspy.core.Stream`

        """

        trace_list = []
        for channel in self.channels:
            if channel[0] in ["e"]:
                ch_type = "electric"
            elif channel[0] in ["h", "b"]:
                ch_type = "magnetic"
            else:
                ch_type = "auxiliary"
            ts_obj = ChannelTS(ch_type, self.dataset[channel])
            trace_list.append(ts_obj.to_obspy_trace())

        return Stream(traces=trace_list)

    def from_obspy_stream(self, obspy_stream, run_metadata=None):
        """
        Get a run from an :class:`obspy.core.stream` which is a list of
        :class:`obspy.core.Trace` objects.

        :param obspy_stream: Obspy Stream object
        :type obspy_stream: :class:`obspy.core.Stream`


        """

        if not isinstance(obspy_stream, Stream):
            msg = f"Input must be obspy.core.Stream not {type(obspy_stream)}"
            self.logger.error(msg)
            raise TypeError(msg)

        array_list = []
        station_list = []
        for obs_trace in obspy_stream:
            channel_ts = ChannelTS()
            channel_ts.from_obspy_trace(obs_trace)
            if channel_ts.channel_metadata.component == "e1":
                channel_ts.channel_metadata.component = "ex"
            if channel_ts.channel_metadata.component == "e2":
                channel_ts.channel_metadata.component = "ey"
            if channel_ts.channel_metadata.component == "h1":
                channel_ts.channel_metadata.component = "hx"
            if channel_ts.channel_metadata.component == "h2":
                channel_ts.channel_metadata.component = "hy"
            if channel_ts.channel_metadata.component == "h3":
                channel_ts.channel_metadata.component = "hz"
            if run_metadata:
                try:
                    ch = [
                        ch
                        for ch in run_metadata.channels
                        if ch.component == channel_ts.component
                    ][0]
                    channel_ts.channel_metadata.update(ch)
                except IndexError:
                    self.logger.warning(
                        "could not find %s" % channel_ts.component
                    )
            station_list.append(channel_ts.station_metadata.fdsn.id)

            array_list.append(channel_ts)

        ### need to merge metadata into something useful, station name is the only
        ### name that is preserved
        try:
            station = list(set([ss for ss in station_list if ss is not None]))[
                0
            ]
        except IndexError:
            station = None
            msg = "Could not find station name"
            self.logger.warn(msg)

        self.station_metadata.fdsn.id = station

        self.set_dataset(array_list)

        # need to be sure update any input metadata.
        if run_metadata is not None:
            self.station_metadata.runs = ListDict()
            self.station_metadata.add_run(run_metadata)

        self.validate_metadata()

    def get_slice(self, start, end=None, n_samples=None):
        """

        :param start: DESCRIPTION
        :type start: TYPE
        :param end: DESCRIPTION, defaults to None
        :type end: TYPE, optional
        :param n_samples: DESCRIPTION, defaults to None
        :type n_samples: TYPE, optional
        :raises ValueError: DESCRIPTION
        :return: DESCRIPTION
        :rtype: TYPE

        """
        """
        Get just a chunk of data from the run, this will attempt to find the
        closest points to the given parameters.  
        
        .. note:: We use pandas `slice_indexer` because xarray slice does not
        seem to work as well, even though they should be based on the same 
        code.

        :param start: start time of the slice
        :type start: string or :class:`mt_metadata.utils.mttime.MTime`
        :param end: end time of the slice, defaults to None
        :type end: string or :class:`mt_metadata.utils.mttime.MTime`, optional
        :param n_samples: number of samples to get, defaults to None
        :type n_samples: int, optional
        :raises ValueError: If end and n_samples are not input
        :return: slice of data requested
        :rtype: :class:`mth5.timeseries.RunTS`

        """
        if not isinstance(start, MTime):
            start = MTime(start)

        if n_samples is not None:
            seconds = (n_samples - 1) / self.sample_rate
            end = start + seconds

        elif end is not None:
            if not isinstance(end, MTime):
                end = MTime(end)
        else:
            raise ValueError("Must input n_samples or end")

        chunk = self.dataset.indexes["time"].slice_indexer(
            start=np.datetime64(start.iso_no_tz),
            end=np.datetime64(end.iso_no_tz),
        )

        new_runts = RunTS()
        new_runts.station_metadata = self.station_metadata
        new_runts.run_metadata = self.run_metadata
        new_runts.filters = self.filters
        new_runts.dataset = self._dataset.isel(indexers={"time": chunk})

        return new_runts

    def calibrate(self, **kwargs):
        """
        Calibrate the data according to the filters in each channel.

        :return: calibrated run
        :rtype: :class:`mth5.timeseries.RunTS`

        """

        new_run = RunTS()
        new_run.run_metadata = self.run_metadata
        new_run.station_metadata = self.station_metadata

        for channel in self.channels:
            ch_ts = getattr(self, channel)
            calibrated_ch_ts = ch_ts.remove_instrument_response(**kwargs)
            new_run.add_channel(calibrated_ch_ts)

        return new_run

    def plot(
        self,
        color_map={
            "ex": (1, 0.2, 0.2),
            "ey": (1, 0.5, 0),
            "hx": (0, 0.5, 1),
            "hy": (0.5, 0.2, 1),
            "hz": (0.2, 1, 1),
        },
        channel_order=None,
    ):
        """

        plot the time series probably slow for large data sets

        """

        if channel_order is not None:
            ch_list = channel_order()
        else:
            ch_list = self.channels

        n_channels = len(self.channels)

        fig = plt.figure()
        fig.subplots_adjust(hspace=0)
        ax_list = []
        for ii, comp in enumerate(ch_list, 1):
            try:
                color = color_map[comp]
            except KeyError:
                color = (0, 0.4, 0.8)
            if ii == 1:
                ax = plt.subplot(n_channels, 1, ii)
            else:
                ax = plt.subplot(n_channels, 1, ii, sharex=ax_list[0])
            self.dataset[comp].plot.line(ax=ax, color=color)
            ax.grid(which="major", color=(0.65, 0.65, 0.65), ls="--", lw=0.75)
            ax.grid(which="minor", color=(0.85, 0.85, 0.85), ls="--", lw=0.5)
            ax.set_axisbelow(True)
            if ii != len(ch_list):
                plt.setp(ax.get_xticklabels(), visible=False)

            ax_list.append(ax)



# -*- coding: utf-8 -*-
"""

Created on Fri Jun 10 07:52:03 2022

:author: Jared Peacock

:license: MIT

"""

# =============================================================================
# Imports
# =============================================================================
import json
from pathlib import Path
from types import SimpleNamespace

from mt_metadata.timeseries import Survey, Station, Run, Electric, Magnetic

# =============================================================================


def read_json_to_object(fn):
    """
    read a json file directly into an object

    :param fn: DESCRIPTION
    :type fn: TYPE
    :return: DESCRIPTION
    :rtype: TYPE

    """

    with open(fn, "r") as fid:
        obj = json.load(fid, object_hook=lambda d: SimpleNamespace(**d))
    return obj


class ConfigJSON:
    """
    A container for the config.json file used to control the recording

    """

    def __init__(self, fn=None, **kwargs):

        self.fn = fn
        self.obj = None

    @property
    def fn(self):
        return self._fn

    @fn.setter
    def fn(self, fn):
        if fn is None:
            self._fn = None
        else:
            fn = Path(fn)
            if fn.exists():
                self._fn = Path(fn)
            else:
                raise ValueError(f"Could not find {fn}")

    def read(self, fn=None):
        """
        read a config.json file that is in the Phoenix format

        :param fn: DESCRIPTION, defaults to None
        :type fn: TYPE, optional
        :return: DESCRIPTION
        :rtype: TYPE

        """

        if fn is not None:
            self.fn = fn
        self.obj = read_json_to_object(self.fn)

    def has_obj(self):
        if self.obj is not None:
            return True
        return False

    @property
    def auto_power_enabled(self):
        if self.has_obj():
            return self.obj.auto_power_enabled

    @property
    def config(self):
        if self.has_obj():
            return self.obj.config[0]

    @property
    def empower_version(self):
        if self.has_obj():
            return self.obj.empower_version

    @property
    def mtc150_reset(self):
        if self.has_obj():
            return self.obj.mtc150_reset

    @property
    def network(self):
        if self.has_obj():
            return self.obj.network

    @property
    def receiver(self):
        if self.has_obj():
            return self.obj.receiver

    @property
    def schedule(self):
        if self.has_obj():
            return self.obj.schedule

    @property
    def surveyTechnique(self):
        if self.has_obj():
            return self.obj.surveyTechnique

    @property
    def timezone(self):
        if self.has_obj():
            return self.obj.timezone

    @property
    def timezone_offset(self):
        if self.has_obj():
            return self.obj.timezone_offset

    @property
    def version(self):
        if self.has_obj():
            return self.obj.version

    def station_metadata(self):
        s = Station()

        s.id = self.config.layout.Station_Name
        s.acquired_by.name = self.config.layout.Operator
        s.acquired_by.organization = self.config.layout.Company_Name
        s.comments = self.config.layout.Notes

        return s


class ReceiverMetadataJSON:
    """
    A container for the recmeta.json file used to control the recording

    """

    def __init__(self, fn=None, **kwargs):

        self.fn = fn
        self.obj = None

        self._e_map = {
            "tag": "component",
            "ty": "type",
            "ga": "gain",
            "sampleRate": "sample_rate",
            "pot_p": "contact_resistance.start",
            "pot_n": "contact_resistance.end",
        }

        self._h_map = {
            "tag": "component",
            "ty": "type",
            "ga": "gain",
            "sampleRate": "sample_rate",
            "type_name": "sensor.model",
            "type": "sensor.type",
            "serial": "sensor.id",
        }

        if self.fn is not None:
            self.read()

    @property
    def fn(self):
        return self._fn

    @fn.setter
    def fn(self, fn):
        if fn is None:
            self._fn = None
        else:
            fn = Path(fn)
            if fn.exists():
                self._fn = Path(fn)
            else:
                raise ValueError(f"Could not find {fn}")

    def read(self, fn=None):
        """
        read a config.json file that is in the Phoenix format

        :param fn: DESCRIPTION, defaults to None
        :type fn: TYPE, optional
        :return: DESCRIPTION
        :rtype: TYPE

        """

        if fn is not None:
            self.fn = fn
        self.obj = read_json_to_object(self.fn)

    def has_obj(self):
        if self.obj is not None:
            return True
        return False

    @property
    def channel_map(self):
        return dict([(d.idx, d.tag) for d in self.obj.channel_map.mapping])

    def get_ch_index(self, tag):
        if self.has_obj():
            for item in self.obj.channel_map.mapping:
                if item.tag.lower() == tag.lower():
                    return item.idx
            raise ValueError(f"Could not find {tag} in channel map.")

    def get_ch_tag(self, index):
        if self.has_obj():
            for item in self.obj.channel_map.mapping:
                if item.idx == index:
                    return item.tag
            raise ValueError(f"Could not find {index} in channel map.")

    def _to_electric_metadata(self, tag):
        c = Electric()

        if self.has_obj():
            ch = self.obj.chconfig.chans[self.get_ch_index(tag)]

            for p_key, m_value in self._e_map.items():
                if p_key == "ty":
                    m_value = "electric"
                c.set_attr_from_name(m_value, getattr(ch, p_key))
            c.channel_number = self.get_ch_index(tag)
            c.dipole_length = ch.length1 + ch.length2
            c.units = "millivolts"
            c.time_period.start = self.obj.start
            c.time_period.end = self.obj.stop
        return c

    def _to_magnetic_metadata(self, tag):
        c = Magnetic()

        if self.has_obj():
            ch = self.obj.chconfig.chans[self.get_ch_index(tag)]

            for p_key, m_value in self._h_map.items():
                if p_key == "ty":
                    m_value = "magnetic"
                c.set_attr_from_name(m_value, getattr(ch, p_key))
            c.channel_number = self.get_ch_index(tag)
            c.sensor.manufacturer = "Phoenix Geophysics"
            c.units = "millivolts"
            c.time_period.start = self.obj.start
            c.time_period.end = self.obj.stop
        return c

    ### should think about putting this part in set_attr
    @property
    def e1_metadata(self):
        return self._to_electric_metadata("e1")

    @property
    def e2_metadata(self):
        return self._to_electric_metadata("e2")

    @property
    def h1_metadata(self):
        return self._to_magnetic_metadata("h1")

    @property
    def h2_metadata(self):
        return self._to_magnetic_metadata("h2")

    @property
    def h3_metadata(self):
        return self._to_magnetic_metadata("h3")

    @property
    def h4_metadata(self):
        return self._to_magnetic_metadata("h4")

    @property
    def h5_metadata(self):
        return self._to_magnetic_metadata("h5")

    @property
    def h6_metadata(self):
        return self._to_magnetic_metadata("h6")

    def get_ch_metadata(self, index):
        """
        get channel metadata from index
        """

        tag = self.get_ch_tag(index)

        return getattr(self, f"{tag.lower()}_metadata")

    ### need to add station and run metadata objects and should be good to go.
    @property
    def run_metadata(self):
        r = Run()
        if self.has_obj():
            r.data_logger.type = self.obj.receiver_model
            r.data_logger.model = self.obj.receiver_commercial_name
            r.data_logger.firmware.version = self.obj.motherboard.mb_fw_ver
            r.data_logger.timing_system.drift = self.obj.timing.tm_drift
        return r

    @property
    def station_metadata(self):
        s = Station()
        if self.has_obj():
            s.id = self.obj.layout.Station_Name
            s.comments = self.obj.layout.Notes
            try:
                s.acquired_by.organization = self.obj.layout.Company_Name
            except AttributeError:
                pass

            s.acquired_by.name = self.obj.layout.Operator
            s.location.latitude = self.obj.timing.gps_lat
            s.location.longitude = self.obj.timing.gps_lon
            s.location.elevation = self.obj.timing.gps_alt
        return s

    @property
    def survey_metadata(self):
        s = Survey()
        if self.has_obj():
            s.id = self.obj.layout.Survey_Name
        return s



from .readers import ReceiverMetadataJSON, ConfigJSON
from .read import read_phoenix, open_phoenix
from .phoenix_collection import PhoenixCollection


__all__ = [
    "ReceiverMetadataJSON",
    "ConfigJSON",
    "read_phoenix",
    "open_phoenix",
    "PhoenixCollection",
]



# -*- coding: utf-8 -*-
"""
Created on Wed Dec 23 17:08:40 2020

Need to make a group for FAP and FIR filters.


:copyright: 
    Jared Peacock (jpeacock@usgs.gov)

:license: MIT

"""

# =============================================================================
# Imports
# =============================================================================

from mth5.groups.base import BaseGroup
from mth5.groups.filter_groups import (
    ZPKGroup,
    CoefficientGroup,
    TimeDelayGroup,
    FAPGroup,
    FIRGroup,
)

# =============================================================================
# Filters Group
# =============================================================================


class FiltersGroup(BaseGroup):
    """
    Not implemented yet
    """

    def __init__(self, group, **kwargs):

        super().__init__(group, **kwargs)

        try:
            self.zpk_group = ZPKGroup(self.hdf5_group.create_group("zpk"))
        except ValueError:
            self.zpk_group = ZPKGroup(self.hdf5_group["zpk"])

        try:
            self.coefficient_group = CoefficientGroup(
                self.hdf5_group.create_group("coefficient")
            )
        except ValueError:
            self.coefficient_group = CoefficientGroup(self.hdf5_group["coefficient"])

        try:
            self.time_delay_group = TimeDelayGroup(
                self.hdf5_group.create_group("time_delay")
            )
        except ValueError:
            self.time_delay_group = TimeDelayGroup(self.hdf5_group["time_delay"])

        try:
            self.fap_group = FAPGroup(self.hdf5_group.create_group("fap"))
        except ValueError:
            self.fap_group = FAPGroup(self.hdf5_group["fap"])

        try:
            self.fir_group = FIRGroup(self.hdf5_group.create_group("fir"))
        except ValueError:
            self.fir_group = FIRGroup(self.hdf5_group["fir"])

    @property
    def filter_dict(self):
        filter_dict = {}
        filter_dict.update(self.zpk_group.filter_dict)
        filter_dict.update(self.coefficient_group.filter_dict)
        filter_dict.update(self.time_delay_group.filter_dict)
        filter_dict.update(self.fap_group.filter_dict)
        filter_dict.update(self.fir_group.filter_dict)

        return filter_dict

    def add_filter(self, filter_object):
        """
        Add a filter dataset based on type

        current types are:
            * zpk         -->  zeros, poles, gain
            * fap         -->  frequency look up table
            * time_delay  -->  time delay filter
            * coefficient -->  coefficient filter

        :param filter_object: An MT metadata filter object
        :type filter_object: :class:`mt_metadata.timeseries.filters`

        """
        filter_object.name = filter_object.name.replace("/", " per ")

        if filter_object.type in ["zpk", "poles_zeros"]:
            try:
                return self.zpk_group.from_object(filter_object)
            except ValueError:
                self.logger.debug("group %s already exists", filter_object.name)
                return self.zpk_group.get_filter(filter_object.name)

        elif filter_object.type in ["coefficient"]:
            try:
                return self.coefficient_group.from_object(filter_object)
            except ValueError:
                self.logger.debug("group %s already exists", filter_object.name)
                return self.coefficient_group.get_filter(filter_object.name)

        elif filter_object.type in ["time_delay", "time delay"]:
            try:
                return self.time_delay_group.from_object(filter_object)
            except ValueError:
                self.logger.debug("group %s already exists", filter_object.name)
                return self.time_delay_group.get_filter(filter_object.name)

        elif filter_object.type in ["fap", "frequency response table"]:
            try:
                return self.fap_group.from_object(filter_object)
            except ValueError:
                self.logger.debug("group %s already exists", filter_object.name)
                return self.fap_group.get_filter(filter_object.name)

        elif filter_object.type in ["fir"]:
            try:
                return self.fir_group.from_object(filter_object)
            except ValueError:
                self.logger.debug("group %s already exists", filter_object.name)
                return self.fir_group.get_filter(filter_object.name)

    def get_filter(self, name):
        """
        Get a filter by name
        """

        try:
            hdf5_ref = self.filter_dict[name]["hdf5_ref"]
        except KeyError:
            msg = "Could not find %s in the filter dictionary"
            self.logger.error(msg, name)
            raise KeyError(msg, name)

        return self.hdf5_group[hdf5_ref]

    def to_filter_object(self, name):
        """
        return the MT metadata representation of the filter
        """

        try:
            f_type = self.filter_dict[name]["type"]
        except KeyError:
            msg = "Could not find %s in the filter dictionary"
            self.logger.error(msg, name)
            raise KeyError(msg, name)

        if f_type in ["zpk"]:
            return self.zpk_group.to_object(name)
        elif f_type in ["coefficient"]:
            return self.coefficient_group.to_object(name)
        elif f_type in ["time_delay", "time delay"]:
            return self.time_delay_group.to_object(name)
        elif f_type in ["fap", "frequency response table"]:
            return self.fap_group.to_object(name)
        elif f_type in ["fir"]:
            return self.fir_group.to_object(name)



from .z3d_header import Z3DHeader
from .z3d_schedule import Z3DSchedule
from .z3d_metadata import Z3DMetadata


from .zen import Z3D, read_z3d
from .coil_response import CoilResponse
from .z3d_collection import Z3DCollection


__all__ = [
    "Z3DHeader",
    "Z3DSchedule",
    "Z3DMetadata",
    "Z3D",
    "CoilResponse",
    "read_z3d",
    "Z3DCollection",
]



# -*- coding: utf-8 -*-
"""
Created on Thu Mar 10 09:02:16 2022

@author: jpeacock
"""

# =============================================================================
# Imports
# =============================================================================
import weakref

import h5py
import numpy as np
import xarray as xr

from mt_metadata.transfer_functions.tf import StatisticalEstimate

from mth5.utils.exceptions import MTH5Error
from mth5.helpers import to_numpy_type
from mth5.utils.mth5_logger import setup_logger

# =============================================================================


class EstimateDataset:
    """
    Holds a statistical estimate

    This will hold multi-dimensional statistical estimates for transfer
    functions.

    :param dataset: hdf5 dataset
    :type dataset: h5py.Dataset
    :param dataset_metadata: data set metadata see
    :class:`mt_metadata.transfer_functions.tf.StatisticalEstimate`,
     defaults to None
    :type dataset_metadata: :class:`mt_metadata.transfer_functions.tf.StatisticalEstimate`, optional
    :param write_metadata: True to write metadata, defaults to True
    :type write_metadata: Boolean, optional
    :param **kwargs: DESCRIPTION
    :type **kwargs: TYPE
    :raises MTH5Error: When an estimate is not present, or metadata name
     does not match the given name

    """

    def __init__(
        self, dataset, dataset_metadata=None, write_metadata=True, **kwargs
    ):

        if dataset is not None and isinstance(dataset, (h5py.Dataset)):
            self.hdf5_dataset = weakref.ref(dataset)()

        self.logger = setup_logger(f"{__name__}.{self._class_name}")

        # set metadata to the appropriate class.  Standards is not a
        # Base object so should be skipped. If the class name is not
        # defined yet set to Base class.
        self.metadata = StatisticalEstimate()

        if not hasattr(self.metadata, "mth5_type"):
            self._add_base_attributes()
            self.metadata.hdf5_reference = self.hdf5_dataset.ref
            self.metadata.mth5_type = self._class_name

        # if the input data set already has filled attributes, namely if the
        # channel data already exists then read them in with our writing back
        if "mth5_type" in list(self.hdf5_dataset.attrs.keys()):
            self.metadata.from_dict(
                {self.hdf5_dataset.attrs["mth5_type"]: self.hdf5_dataset.attrs}
            )

        # if metadata is input, make sure that its the same class type amd write
        # to the hdf5 dataset
        if dataset_metadata is not None:
            if not isinstance(dataset_metadata, type(self.metadata)):
                msg = "metadata must be type metadata.%s not %s"
                self.logger.error(
                    msg, self._class_name, type(dataset_metadata)
                )
                raise MTH5Error(msg % self._class_name, type(dataset_metadata))

            # load from dict because of the extra attributes for MTH5
            self.metadata.from_dict(dataset_metadata.to_dict())
            self.metadata.hdf5_reference = self.hdf5_dataset.ref
            self.metadata.mth5_type = self._class_name

            # write out metadata to make sure that its in the file.
            if write_metadata:
                self.write_metadata()

        # if the attrs don't have the proper metadata keys yet write them
        if not "mth5_type" in list(self.hdf5_dataset.attrs.keys()):
            self.write_metadata()

    def _add_base_attributes(self):
        # add 2 attributes that will help with querying
        # 1) the metadata class name
        self.metadata.add_base_attribute(
            "mth5_type",
            self._class_name,
            {
                "type": str,
                "required": True,
                "style": "free form",
                "description": "type of group",
                "units": None,
                "options": [],
                "alias": [],
                "example": "group_name",
                "default": None,
            },
        )

        # 2) the HDF5 reference that can be used instead of paths
        self.metadata.add_base_attribute(
            "hdf5_reference",
            self.hdf5_dataset.ref,
            {
                "type": "h5py_reference",
                "required": True,
                "style": "free form",
                "description": "hdf5 internal reference",
                "units": None,
                "options": [],
                "alias": [],
                "example": "<HDF5 Group Reference>",
                "default": None,
            },
        )

    def __str__(self):
        return self.metadata.to_json()

    def __repr__(self):
        return self.__str__()

    @property
    def _class_name(self):
        return self.__class__.__name__.split("Dataset")[0]

    def read_metadata(self):
        """
        Read metadata from the HDF5 file into the metadata container, that
        way it can be validated.

        """

        self.metadata.from_dict({self._class_name: self.hdf5_dataset.attrs})

    def write_metadata(self):
        """
        Write metadata from the metadata container to the HDF5 attrs
        dictionary.

        """
        meta_dict = self.metadata.to_dict()[self.metadata._class_name.lower()]
        for key, value in meta_dict.items():
            value = to_numpy_type(value)
            self.hdf5_dataset.attrs.create(key, value)

    def replace_dataset(self, new_data_array):
        """
        replace the entire dataset with a new one, nothing left behind

        :param new_data_array: new data array
        :type new_data_array: :class:`numpy.ndarray`

        """
        if not isinstance(new_data_array, np.ndarray):
            try:
                new_data_array = np.array(new_data_array)
            except (ValueError, TypeError) as error:
                msg = f"{error} Input must be a numpy array not {type(new_data_array)}"
                self.logger.exception(msg)
                raise TypeError(msg)

        if new_data_array.shape != self.hdf5_dataset.shape:
            self.hdf5_dataset.resize(new_data_array.shape)

        self.hdf5_dataset[...] = new_data_array

    def to_xarray(self, period):
        """
        :return: an xarray DataArray with appropriate metadata and the
         appropriate coordinates.
        :rtype: :class:`xarray.DataArray`

        .. note:: that metadta will not be validated if changed in an xarray.

        loads from memory
        """

        return xr.DataArray(
            data=self.hdf5_dataset[()],
            dims=["period", "output", "input"],
            name=self.metadata.name,
            coords=[
                ("period", period),
                ("output", self.metadata.output_channels),
                ("input", self.metadata.input_channels),
            ],
            attrs=self.metadata.to_dict(single=True),
        )

    def to_numpy(self):
        """
        :return: a numpy structured array with
        :rtype: :class:`numpy.ndarray`

        loads into RAM

        """

        return self.hdf5_dataset[()]

    def from_numpy(self, new_estimate):
        """
        :return: a numpy structured array
        :rtype: :class:`numpy.ndarray`

        .. note:: data is a builtin to numpy and cannot be used as a name

        loads into RAM

        """

        if not isinstance(new_estimate, np.ndarray):
            try:
                new_estimate = np.array(new_estimate)
            except (ValueError, TypeError) as error:
                msg = f"{error} Input must be a numpy array not {type(new_estimate)}"
                self.logger.exception(msg)
                raise TypeError(msg)

        if new_estimate.dtype != self.hdf5_dataset.dtype:
            msg = "Input array must be type %s not %s"
            self.logger.error(msg, new_estimate.dtype, self.hdf5_dataset.dtype)
            raise TypeError(
                msg % (new_estimate.dtype, self.hdf5_dataset.dtype)
            )

        if new_estimate.shape != self.hdf5_dataset.shape:
            self.hdf5_dataset.resize(new_estimate.shape)

        self.hdf5_dataset[...] = new_estimate

    def from_xarray(self, data):
        """
        :return: an xarray DataArray with appropriate metadata and the
         appropriate coordinates base on the metadata.
        :rtype: :class:`xarray.DataArray`

        .. note:: that metadta will not be validated if changed in an xarray.

        loads from memory
        """

        self.metadata.output_channels = data.coords["output"].values.tolist()
        self.metadata.input_channels = data.coords["input"].values.tolist()
        self.metadata.name = data.name
        self.metadata.data_type = data.dtype.name

        self.write_metadata()

        self.from_numpy(data.to_numpy())



# -*- coding: utf-8 -*-
"""
LEMI 424 Collection
====================

Collection of TXT files combined into runs

Created on Wed Aug 31 10:32:44 2022

@author: jpeacock
"""

# =============================================================================
# Imports
# =============================================================================
import pandas as pd

from mth5.io.collection import Collection
from mth5.io.lemi import LEMI424

# =============================================================================


class LEMICollection(Collection):
    """
    Collection of LEMI 424 files into runs based on start and end times.
    Will assign the run name as 'sr1_{index:0{zeros}}' --> 'sr1_0001' for
    `zeros` = 4.

    :param file_path: full path to single station LEMI424 directory
    :type file_path: string or :class`pathlib.Path`
    :param file_ext: extension of LEMI424 files, default is 'txt'
    :type file_ext: string
    :param station_id: station id
    :type station_id: string
    :param survey_id: survey id
    :type survey_id: string

    .. note:: This class assumes that the given file path contains a single
     LEMI station.  If you want to do multiple stations merge the returned
     data frames.

    .. note:: LEMI data comes with little metadata about the station or survey,
     therefore you should assign `station_id` and `survey_id`.

    .. code-block:: python

        >>> from mth5.io.lemi import LEMICollection
        >>> lc = LEMICollection(r"/path/to/single/lemi/station")
        >>> lc.station_id = "mt001"
        >>> lc.survey_id = "test_survey"
        >>> run_dict = lc.get_runs(1)


    """

    def __init__(self, file_path=None, **kwargs):
        super().__init__(file_path=file_path, **kwargs)
        self.file_ext = "txt"

        self.station_id = "mt001"
        self.survey_id = "mt"

    def to_dataframe(
        self, sample_rates=[1], run_name_zeros=4, calibration_path=None
    ):
        """
        Create a data frame of each TXT file in a given directory.

        .. note:: This assumes the given directory contains a single station

        :param sample_rates: sample rate to get, will always be 1 for LEMI data
         defaults to [1]
        :type sample_rates: int or list, optional
        :param run_name_zeros: number of zeros to assing to the run name,
         defaults to 4
        :type run_name_zeros: int, optional
        :param calibration_path: path to calibration files, defaults to None
        :type calibration_path: string or Path, optional
        :return: Dataframe with information of each TXT file in the given
         directory.
        :rtype: :class:`pandas.DataFrame`

        :Example:

            >>> from mth5.io.lemi import LEMICollection
            >>> lc = LEMICollection("/path/to/single/lemi/station")
            >>> lemi_df = lc.to_dataframe()

        """

        entries = []
        for fn in self.get_files(self.file_ext):
            lemi_obj = LEMI424(fn)
            n_samples = int(lemi_obj.n_samples)
            lemi_obj.read_metadata()

            entry = {}
            entry["survey"] = self.survey_id
            entry["station"] = self.station_id
            entry["run"] = None
            entry["start"] = lemi_obj.start.isoformat()
            entry["end"] = lemi_obj.end.isoformat()
            entry["channel_id"] = 1
            entry["component"] = ",".join(
                lemi_obj.run_metadata.channels_recorded_all
            )
            entry["fn"] = fn
            entry["sample_rate"] = lemi_obj.sample_rate
            entry["file_size"] = lemi_obj.file_size
            entry["n_samples"] = n_samples
            entry["sequence_number"] = 0
            entry["instrument_id"] = "LEMI424"
            entry["calibration_fn"] = None

            entries.append(entry)

        # make pandas dataframe and set data types
        df = self._sort_df(
            self._set_df_dtypes(pd.DataFrame(entries)), run_name_zeros
        )

        return df

    def assign_run_names(self, df, zeros=4):
        """
        Assign run names based on start and end times, checks if a file has
        the same start time as the last end time.

        Run names are assigned as sr{sample_rate}_{run_number:0{zeros}}.

        :param df: Dataframe with the appropriate columns
        :type df: :class:`pandas.DataFrame`
        :param zeros: number of zeros in run name, defaults to 4
        :type zeros: int, optional
        :return: Dataframe with run names
        :rtype: :class:`pandas.DataFrame`

        """
        count = 1
        for row in df.itertuples():
            if row.Index == 0:
                df.loc[row.Index, "run"] = f"sr1_{count:0{zeros}}"
                previous_end = row.end
            else:
                if (
                    row.start - previous_end
                ).total_seconds() / row.sample_rate == row.sample_rate:
                    df.loc[row.Index, "run"] = f"sr1_{count:0{zeros}}"
                else:
                    count += 1
                    df.loc[row.Index, "run"] = f"sr1_{count:0{zeros}}"
                previous_end = row.end

        return df



#!/usr/bin/env python

"""
time series filters

"""

# =================================================================
import numpy as np
from scipy import signal

from matplotlib import pyplot as plt
from matplotlib.lines import Line2D
from mth5.utils.mth5_logger import setup_logger

logger = setup_logger(__file__)
# =================================================================


def butter_bandpass(lowcut, highcut, fs, order=5):
    """
    Butterworth bandpass filter using scipy.signal

    :param lowcut: low cut frequency in Hz
    :type lowcut: float
    :param highcut: high cut frequency in Hz
    :type highcut: float
    :param fs: Sample rate
    :type fs: float
    :param order: Butterworth order, defaults to 5
    :type order: int, optional
    :return: SOS scipy.signal format
    :rtype: scipy.signal.SOS?

    """
    nyq = 0.5 * fs

    if lowcut is not None:
        low = lowcut / nyq
    if highcut is not None:
        high = highcut / nyq
    if lowcut and highcut:
        sos = signal.butter(
            order, [low, high], analog=False, btype="band", output="sos"
        )
    elif highcut is None:
        sos = signal.butter(order, low, analog=False, btype="low", output="sos")
    elif lowcut is None:
        sos = signal.butter(
            order, high, analog=False, btype="high", output="sos"
        )
    return sos


def butter_bandpass_filter(data, lowcut, highcut, fs, order=5):
    """

    :param data: 1D time series data
    :type data: np.ndarray
    :param lowcut: low cut frequency in Hz
    :type lowcut: float
    :param highcut: high cut frequency in Hz
    :type highcut: float
    :param fs: Sample rate
    :type fs: float
    :param order: Butterworth order, defaults to 5
    :type order: int, optional
    :return: filtered data
    :rtype: np.ndarray

    """
    sos = butter_bandpass(lowcut, highcut, fs, order=order)
    y = signal.sosfiltfilt(sos, data)
    return y


def low_pass(data, low_pass_freq, cutoff_freq, sampling_rate):
    """

    :param data: 1D time series data
    :type data: np.ndarray
    :param low_pass_freq: low pass frequency in Hz
    :type low_pass_freq: float
    :param cutoff_freq: cut off frequency in Hz
    :type cutoff_freq: float
    :param sampling_rate: Sample rate in samples per second
    :type sampling_rate: float
    :return: lowpass filtered data
    :rtype: np.ndarray

    """
    nyq = 0.5 * sampling_rate
    filt_order, wn = signal.buttord(
        low_pass_freq / nyq, cutoff_freq / nyq, 3, 40
    )

    b, a = signal.butter(filt_order, wn, btype="low")
    data_filtered = signal.filtfilt(b, a, data)

    return data_filtered


def zero_pad(input_array, power=2, pad_fill=0):
    """

    :param input_array: 1D array
    :type input_array: np.ndarray
    :param power: base power to used to pad to, defaults to 2 which is optimal
    for the FFT
    :type power: int, optional
    :param pad_fill: fill value for padded values, defaults to 0
    :type pad_fill: float, optional
    :return: zero padded array
    :rtype: np.ndarray

    """

    len_array = input_array.shape[0]
    if power == 2:
        npow = int(np.ceil(np.log2(len_array)))
    if power == 10:
        npow = int(np.ceil(np.log10(len_array)))
    if npow > 32:
        logger.warning(
            "Exceeding memory allocation inherent in your computer 2**32. "
            "Limiting the zero pad to 2**32"
        )
    pad_array = np.zeros(power**npow)
    if pad_fill != 0:
        pad_array[:] = pad_fill
    pad_array[0:len_array] = input_array

    return pad_array


class RemoveInstrumentResponse:
    """
    Remove instrument response from the given channel response filter

    The order of operations is important (if applied):

        1) detrend
        2) zero mean
        3) zero pad
        4) time window
        5) frequency window
        6) remove response
        7) undo time window
        8) bandpass

    :param ts: time series data to remove response from
    :type ts: np.ndarray((N,) , dtype=float)
    :param time_array: time index that corresponds to the time series
    :type time_array: np.ndarray((N,) , dtype=np.datetime[ns])
    :param sample_interval: seconds per sample (time interval between samples)
    :type sample_interval: float
    :param channel_response_filter: Channel response filter with all filters
    included to convert from counts to physical units
    :type channel_response_filter: `class`:mt_metadata.timeseries.filters.ChannelResponseFilter`

    **kwargs**

    :param plot: to plot the calibration process [ False | True ]
    :type plot: boolean, default True
    :param detrend: Remove linar trend of the time series
    :type detrend: boolean, default True
    :param zero_mean: Remove the mean of the time series
    :type zero_mean: boolean, default True
    :param zero_pad: pad the time series to the next power of 2 for efficiency
    :type zero_pad: boolean, default True
    :param t_window: Time domain windown name see `scipy.signal.windows` for options
    :type t_window: string, default None
    :param t_window_params: Time domain window parameters, parameters can be
    found in `scipy.signal.windows`
    :type t_window_params: dictionary
    :param f_window: Frequency domain windown name see `scipy.signal.windows` for options
    :type f_window: string, defualt None
    :param f_window_params: Frequency window parameters, parameters can be
    found in `scipy.signal.windows`
    :type f_window_params: dictionary
    :param bandpass: bandpass freequency and order {"low":, "high":, "order":,}
    :type bandpass: dictionary


    """

    def __init__(
        self,
        ts,
        time_array,
        sample_interval,
        channel_response_filter,
        **kwargs,
    ):
        self.logger = setup_logger(f"{__name__}.{self.__class__.__name__}")
        self.ts = ts
        self.time_array = time_array
        self.sample_interval = sample_interval
        self.channel_response_filter = channel_response_filter
        self.plot = False
        self.detrend = True
        self.zero_mean = True
        self.zero_pad = True
        self.t_window = None
        self.t_window_params = {}
        self.f_window = None
        self.f_window_params = {}
        self.bandpass = {}
        self.fig = None
        self.nrows = None
        self.subplot_dict = {}

        for key, value in kwargs.items():
            setattr(self, key, value)

    def _subplots(self, x, y, color, num, label):
        """
        helper function to make subplots for if plotting is desired

        :param x: x array
        :type x: np.ndarray
        :param y: y array
        :type y: np.ndarray
        :param color: color of the line
        :type color: tuple
        :param num: subplot number
        :type num: integer
        :param label: legend label
        :type label: string

        """
        ax_t = self.fig.get_axes()[0]
        ax_f = self.fig.get_axes()[1]
        ax = self.fig.add_subplot(self.nrows, 2, num, sharex=ax_t)
        ax2 = self.fig.add_subplot(self.nrows, 2, num + 1, sharex=ax_f)

        # plot time series
        line = Line2D([0], [0], color=color)
        ax.plot(x, y, color=color)
        f = np.fft.rfftfreq(x.size, d=self.sample_interval)

        # plot spectra
        data = np.fft.rfft(y)
        ax2.loglog(f, abs(data), color=color)
        ax.legend(
            [line],
            [label],
            loc="upper left",
            borderaxespad=0.01,
            borderpad=0.1,
            markerscale=0.02,
        ).set_zorder(1000)

    @staticmethod
    def get_window(window, window_params, size):
        """
        Get window from scipy.signal

        :param window: name of the window
        :type window: string
        :param window_params: dictionary of window parameters
        :type window_params: dictionary
        :param size: number of points in the window
        :type size: integer
        :return: window function
        :rtype: class:`scipy.signal`

        """
        return getattr(signal.windows, window)(size, **window_params)

    def apply_detrend(self, ts):
        """
        Detrend time series using scipy.detrend('linear')

        :param ts: input time series
        :type ts: np.ndarray
        :return: detrended time series
        :rtype: np.ndarray

        """
        ts = signal.detrend(np.nan_to_num(ts), type="linear")

        if self.plot:
            self._subplots(
                self.time_array,
                ts,
                (0.45, 0.1, 0.5),
                self.subplot_dict["detrend"],
                "Detrended",
            )
        return ts

    def apply_zero_mean(self, ts):
        """
        Remove the mean from the time series

        :param ts: input time series
        :type ts: np.ndarray
        :return: zero mean time series
        :rtype: np.ndarray

        """
        ts = ts - ts.mean()

        if self.plot:
            self._subplots(
                self.time_array,
                ts,
                (0.55, 0.1, 0.4),
                self.subplot_dict["zero_mean"],
                "Zero Mean",
            )
        return ts

    def apply_zero_pad(self, ts):
        """
        zero pad to power of 2, at the end of the time series to make the
        FFT more efficient

        :param ts: input time series
        :type ts: np.ndarray
        :return: zero padded time series
        :rtype: np.ndarray

        """

        pad_ts = zero_pad(ts)

        if self.plot:
            self._subplots(
                self.time_array,
                pad_ts[0 : self.time_array.size],
                (0.7, 0.1, 0.25),
                self.subplot_dict["pad"],
                "Zero Pad",
            )
        return pad_ts

    def apply_t_window(self, ts):
        """
        Apply a window in the time domain. Get the available windows from
        `scipy.signal.windows`

        :param ts: input time series
        :type ts: np.ndarray
        :return: windowed time series
        :rtype: np.ndarray

        """
        w = self.get_window(self.t_window, self.t_window_params, self.ts.size)
        ts = ts * w

        if self.plot:
            self._subplots(
                self.time_array,
                ts,
                (0.7, 0.1, 0.25),
                self.subplot_dict["t_window"],
                f"t_window {self.t_window.capitalize()}",
            )
        return ts

    def apply_f_window(self, data):
        """
        Apply a frequency domain window. Get the available windows from
        `scipy.signal.windows`

        Need to create a window twice the size of the input because we are
        only taking the rfft which gives just half the spectra
        and then take only half the window

        :param data: input spectra
        :type data: np.ndarray
        :return: windowed spectra
        :rtype: np.ndarray

        """

        w = self.get_window(self.f_window, self.f_window_params, 2 * data.size)[
            data.size :
        ]
        data = data * w
        if self.plot:
            f = np.fft.rfftfreq(2 * data.size, d=self.sample_interval)[1:]
            ax_t = self.fig.get_axes()[0]
            ax_f = self.fig.get_axes()[1]
            ax = self.fig.add_subplot(
                self.nrows, 2, self.subplot_dict["f_window"], sharex=ax_t
            )
            ax2 = self.fig.add_subplot(
                self.nrows, 2, self.subplot_dict["f_window"] + 1, sharex=ax_f
            )

            ax.plot(
                self.time_array,
                abs(np.fft.irfft(data)[0 : self.time_array.size]),
                color=(0.85, 0.1, 0.15),
            )
            ax2.loglog(f, abs(data), color=(0.85, 0.1, 0.15))
            axt2 = ax2.twinx()
            axt2.semilogx(f, w, color=(0.5, 0.5, 0.5), zorder=0)
            line = Line2D([0], [0], color=(0.85, 0.1, 0.15))
            ax.legend(
                [line],
                [f"Freq Window {self.f_window.capitalize()}"],
                loc="upper left",
                borderaxespad=0.01,
                borderpad=0.1,
            )
        return data

    def apply_bandpass(self, ts):
        """
        apply a bandpass filter to the calibrated data


        :param ts: calibrated time series
        :type ts: np.ndarray
        :return: bandpassed time series
        :rtype: np.ndarray

        """
        ts = butter_bandpass_filter(
            ts,
            self.bandpass["low"],
            self.bandpass["high"],
            self.sample_interval,
            order=self.bandpass["order"],
        )

        if self.plot:
            self._subplots(
                self.time_array,
                ts,
                (0.875, 0.1, 0.1),
                self.subplot_dict["bandpass"],
                "Band Pass",
            )
        return ts

    def _get_subplot_count(self):
        """
        helper function to get subplot information

        :return: dictionary of subplot information
        :rtype: dictionary

        """
        order = [
            "detrend",
            "zero_mean",
            "t_window",
            "pad",
            "f_window",
            "bandpass",
        ]
        pdict = {
            "pad": self.zero_pad,
            "zero_mean": self.zero_mean,
            "detrend": self.detrend,
            "t_window": self.t_window,
            "f_window": self.f_window,
            "bandpass": self.bandpass,
        }
        subplot_dict = {}
        count = 3
        nrows = 2
        for key in order:
            value = pdict[key]
            if value:
                subplot_dict[key] = count
                count += 2
                nrows += 1
        self.nrows = nrows

        return subplot_dict

    def remove_instrument_response(self, operation="divide"):
        """
        Remove instrument response following the recipe provided

        :return: calibrated time series
        :rtype: np.ndarray

        """
        ts = np.copy(self.ts)
        f = np.fft.rfftfreq(ts.size, d=self.sample_interval)
        step = 1
        if self.plot:
            self.subplot_dict = self._get_subplot_count()
            self.fig = plt.figure(figsize=[10, 12])
            self.fig.clf()
            ax = self.fig.add_subplot(self.nrows, 2, 1)
            ax2 = self.fig.add_subplot(self.nrows, 2, 2)
            (l1,) = ax.plot(self.time_array, ts, color="k", lw=2)
            ax.set_xlim((self.time_array[0], self.time_array[-1]))
            (l2,) = ax2.loglog(f, abs(np.fft.rfft(ts)), "k", lw=2)
            ax2.set_xlim((f[0], f[-1]))
            ax.legend(
                [l1],
                ["Original"],
                loc="upper left",
                borderaxespad=0.01,
                borderpad=0.1,
            )
        # detrend
        if self.detrend:
            ts = self.apply_detrend(ts)
            self.logger.debug(f"Step {step}: Applying Linear Detrend")
            step += 1
        # zero mean
        if self.zero_mean:
            ts = self.apply_zero_mean(ts)
            self.logger.debug(f"Step {step}: Removing Mean")
            step += 1
        # filter in time domain
        if self.t_window is not None:
            ts = self.apply_t_window(ts)
            self.logger.debug(
                f"Step {step}: Applying {self.t_window} Time Window"
            )
            step += 1
            if self.plot:
                wax = self.fig.get_axes()[
                    self.subplot_dict["t_window"] - 1
                ].twinx()
                (tw,) = wax.plot(
                    self.time_array,
                    self.get_window(
                        self.t_window, self.t_window_params, ts.size
                    ),
                    color=(0.75, 0.75, 0.75),
                    zorder=0,
                )
        if self.zero_pad:
            # pad the time series to a power of 2, this may be overkill, especially for long time series
            ts = self.apply_zero_pad(ts)
            self.logger.debug(f"Step {step}: Applying Zero Padding")
            step += 1
        # get the real frequencies of the FFT
        f = np.fft.rfftfreq(ts.size, d=self.sample_interval)

        if self.channel_response_filter.filters_list is []:
            raise ValueError(
                "There are no filters in channel_response to remove"
            )
        # compute the complex response given the frequency range of the FFT
        # the complex response assumes frequencies are in reverse order and flip them on input
        # so we need to flip the complex reponse so it aligns with the fft.
        cr = self.channel_response_filter.complex_response(f)[::-1]
        # remove the DC term at frequency == 0
        cr[-1] = abs(cr[-2]) + 0.0j

        data = np.fft.rfft(ts)
        # remove DC term
        data[-1] = abs(data[-1]) + 0.0j

        # if a window is requested then create it here and mulitply by the data
        # the windows are designed to be symmetrical about frequency = 0
        # here we are taking only the real part of the FFT so we cut the window in half
        if self.f_window is not None:
            data = self.apply_f_window(data)
            self.logger.debug(
                f"Step {step}: Applying {self.f_window} Frequency Window"
            )
            step += 1

        if operation == "divide":
            # calibrate the time series, compute real part of fft, divide out
            # channel response, inverse fft
            calibrated_ts = np.fft.irfft(data / cr)[0 : self.ts.size]
            self.logger.debug(
                f"Step {step}: Removing Calibration by {operation}"
            )
            step += 1

        elif operation == "multiply":
            # calibrate the time series, compute real part of fft, multiply out
            # channel response, inverse fft
            calibrated_ts = np.fft.irfft(data * cr)[0 : self.ts.size]
            self.logger.debug(
                f"Step {step}: Removing Calibration  by {operation}"
            )
            step += 1

        # If a time window was applied, need to un-apply it to reconstruct the signal.
        if self.t_window is not None:
            w = self.get_window(
                self.t_window, self.t_window_params, calibrated_ts.size
            )
            calibrated_ts = calibrated_ts / w
            self.logger.debug(f"Step {step}: Un-applying Time Window")
            step += 1
        if self.bandpass:
            calibrated_ts = self.apply_bandpass(calibrated_ts)
            self.logger.debug(f"Step {step}: Applying Bandpass Filter")
            step += 1
        if self.plot:
            self._subplots(
                self.time_array[0 : calibrated_ts.size],
                calibrated_ts,
                (1, 0.1, 0.1),
                self.nrows * 2 - 1,
                "Calibrated",
            )
            self.fig.get_axes()[-2].set_ylabel(
                self.channel_response_filter.units_in
            )
            if self.t_window is not None:
                wax = self.fig.get_axes()[-2].twinx()
                (tw,) = wax.plot(
                    self.time_array, 1.0 / w, color=(0.5, 0.5, 0.5), zorder=0
                )
            self.fig.tight_layout()
            plt.show()
        return calibrated_ts


def adaptive_notch_filter(
    bx,
    df=100,
    notches=[50, 100],
    notchradius=0.5,
    freqrad=0.9,
    rp=0.1,
    dbstop_limit=5.0,
):
    """

    :param bx: time series to filter
    :type bx: np.ndarray
    :param df: sample rate in samples per second, defaults to 100
    :type df: float, optional
    :param notches: list of frequencies to locate notches at in Hz,
     defaults to [50, 100]
    :type notches: list, optional
    :param notchradius: notch radius, defaults to 0.5
    :type notchradius: float, optional
    :param freqrad: radius to search for a peak at the notch frequency,
     defaults to 0.9
    :type freqrad: float, optional
    :param rp: ripple of Chebyshev type 1 filter, lower numbers means less
     ripples, defaults to 0.1
    :type rp: float, optional
    :param dbstop_limit: limits the difference between the peak at the
     notch and surrounding spectra.  Any difference
     above dbstop_limit will be filtered, anything
     less will not, defaults to 5.0
    :type dbstop_limit: float, optional
    :return: notch filtered data
    :rtype: np.ndarray
    :return: list of notch frequencies
    :rtype: list


    Example
    ---------

        >>> import RemovePeriodicNoise_Kate as rmp
        >>> # make a variable for the file to load in
        >>> fn = r"/home/MT/mt01_20130101_000000.BX"
        >>> # load in file, if the time series is not an ascii file
        >>> # might need to add keywords to np.loadtxt or use another
        >>> # method to read in the file
        >>> bx = np.loadtxt(fn)
        >>> # create a list of frequencies to filter out
        >>> freq_notches = [50, 150, 200]
        >>> # filter data
        >>> bx_filt, filt_lst = rmp.adaptiveNotchFilter(bx, df=100.
        >>> ...                                         notches=freq_notches)
        >>> #save the filtered data into a file
        >>> np.savetxt(r"/home/MT/Filtered/mt01_20130101_000000.BX", bx_filt)

    Notes:

        Most of the time the default parameters work well, the only thing
        you need to change is the notches and perhaps the radius.  I would
        test it out with a few time series to find the optimum parameters.
        Then make a loop over all you time series data. Something like

        >>> import os
        >>> dirpath = r"/home/MT"
        >>> #make a director to save filtered time series
        >>> save_path = r"/home/MT/Filtered"
        >>> if not os.path.exists(save_path):
        >>>     os.mkdir(save_path)
        >>> for fn in os.listdir(dirpath):
        >>>     bx = np.loadtxt(os.path.join(dirpath, fn)
        >>>     bx_filt, filt_lst = rmp.adaptiveNotchFilter(bx, df=100.
        >>>     ...                                         notches=freq_notches)
        >>>     np.savetxt(os.path.join(save_path, fn), bx_filt)

    """

    bx = np.array(bx)

    if type(notches) is list:
        notches = np.array(notches)
    elif type(notches) in [float, int]:
        notches = np.array([notches], dtype=np.float)
    df = float(df)  # make sure df is a float
    dt = 1.0 / df  # sampling rate

    # transform data into frequency domain to find notches
    BX = np.fft.fft(zero_pad(bx))
    n = len(BX)  # length of array
    dfn = df / n  # frequency step
    dfnn = int(freqrad / dfn)  # radius of frequency search
    fn = notchradius  # filter radius
    freq = np.fft.fftfreq(n, dt)

    filtlst = []
    for notch in notches:
        if notch > freq.max():
            break
        else:
            fspot = int(round(notch / dfn))
            nspot = np.where(
                abs(BX)
                == max(abs(BX[max([fspot - dfnn, 0]) : min([fspot + dfnn, n])]))
            )[0][0]

            med_bx = np.median(
                abs(
                    BX[
                        max([nspot - dfnn * 10, 0]) : min(
                            [nspot + dfnn * 10, n]
                        )
                    ]
                )
                ** 2
            )

            # calculate difference between peak and surrounding spectra in dB
            dbstop = 10 * np.log10(abs(BX[nspot]) ** 2 / med_bx)
            if np.nan_to_num(dbstop) == 0.0 or dbstop < dbstop_limit:
                filtlst.append("No need to filter \n")
                pass
            else:
                filtlst.append([freq[nspot], dbstop])
                ws = 2 * np.array([freq[nspot] - fn, freq[nspot] + fn]) / df
                wp = (
                    2
                    * np.array([freq[nspot] - 2 * fn, freq[nspot] + 2 * fn])
                    / df
                )
                ford, wn = signal.cheb1ord(wp, ws, 1, dbstop)
                b, a = signal.cheby1(1, 0.5, wn, btype="bandstop")
                bx = signal.filtfilt(b, a, bx)
    return bx, filtlst



# -*- coding: utf-8 -*-
"""
Created on Wed Mar 23 14:09:38 2022

@author: jpeacock
"""

# =============================================================================
# Imports
# =============================================================================
import pandas as pd
import numpy as np
import h5py

from mth5 import TF_DTYPE
from mth5.tables import MTH5Table
from mth5.helpers import validate_name

# =============================================================================


class TFSummaryTable(MTH5Table):
    """
    Object to hold the channel summary and provide some convenience functions
    like fill, to_dataframe ...
    
    """

    def __init__(self, hdf5_dataset):
        super().__init__(hdf5_dataset)
        self._dtype = TF_DTYPE

    def to_dataframe(self):
        """
        Create a pandas DataFrame from the table for easier querying.
        
        :return: Channel Summary 
        :rtype: :class:`pandas.DataFrame`

        """

        df = pd.DataFrame(self.array[()])
        for key in [
            "station",
            "survey",
            "tf_id",
            "units",
        ]:
            setattr(df, key, getattr(df, key).str.decode("utf-8"))
        return df

    def summarize(self):
        """
        
        :return: DESCRIPTION
        :rtype: TYPE

        """
        self.clear_table()

        def recursive_get_tf_entry(group):
            """
            a function to get tf entry, hopefully this is faster than looping
            and getting the correct group object.
            
            """
            if isinstance(group, (h5py._hl.group.Group, h5py._hl.files.File)):
                for key, node in group.items():
                    try:
                        group_type = node.attrs["mth5_type"].lower()
                        if group_type == "transferfunction":
                            has_impedance = False
                            has_tipper = False
                            has_covariance = False
                            if "transfer_function" in node.keys():
                                tf_dataset = node["transfer_function"]
                                if tf_dataset != (1, 1, 1):
                                    nz = np.nonzero(tf_dataset)
                                    unique_values = np.unique(nz[1])
                                    if 0 in unique_values or 1 in unique_values:
                                        has_impedance = True
                                    if 2 in unique_values:
                                        has_tipper = True
                            if (
                                "residual_covariance" in node.keys()
                                and "inverse_signal_power" in node.keys()
                            ):
                                res = node["residual_covariance"]
                                isp = node["inverse_signal_power"]

                                if res.shape != (1, 1, 1) and isp.shape != (1, 1, 1):
                                    has_covariance = True
                            if "period" in node.keys():
                                period = node["period"][()]
                            else:
                                period = np.zeros(2)
                            tf_entry = np.array(
                                [
                                    (
                                        validate_name(node.parent.parent.attrs["id"]),
                                        validate_name(
                                            node.parent.parent.parent.parent.attrs["id"]
                                        ),
                                        node.parent.parent.attrs["location.latitude"],
                                        node.parent.parent.attrs["location.longitude"],
                                        node.parent.parent.attrs["location.elevation"],
                                        validate_name(node.attrs["id"]),
                                        node.attrs["units"],
                                        has_impedance,
                                        has_tipper,
                                        has_covariance,
                                        period.min(),
                                        period.max(),
                                        node.ref,
                                        node.parent.parent.ref,
                                    )
                                ],
                                dtype=TF_DTYPE,
                            )
                            self.add_row(tf_entry)
                        else:
                            recursive_get_tf_entry(node)
                    except KeyError:
                        recursive_get_tf_entry(node)
            elif isinstance(group, h5py._hl.dataset.Dataset):
                pass

        recursive_get_tf_entry(self.array.parent)



# -*- coding: utf-8 -*-
"""

Tools for FDSN standards

Created on Wed Sep 30 11:47:01 2020

:author: Jared Peacock

:license: MIT

"""

import logging

logger = logging.getLogger(__name__)

period_code_dict = {
    "F": {"min": 1000, "max": 5000},
    "G": {"min": 1000, "max": 5000},
    "D": {"min": 250, "max": 1000},
    "C": {"min": 250, "max": 1000},
    "E": {"min": 80, "max": 250},
    "S": {"min": 10, "max": 80},
    "H": {"min": 80, "max": 250},
    "B": {"min": 10, "max": 80},
    "M": {"min": 1, "max": 10},
    "L": {"min": 0.95, "max": 1.05},
    "V": {"min": 0.095, "max": 0.105},
    "U": {"min": 0.0095, "max": 0.0105},
    "R": {"min": 0.0001, "max": 0.001},
    "P": {"min": 0.00001, "max": 0.0001},
    "T": {"min": 0.000001, "max": 0.00001},
    "Q": {"min": 0, "max": 0.000001},
}

measurement_code_dict = {
    "tilt": "A",
    "creep": "B",
    "calibration": "C",
    "pressure": "D",
    "magnetics": "F",
    "gravity": "G",
    "humidity": "I",
    "temperature": "K",
    "water_current": "O",
    "electric": "Q",
    "rain_fall": "R",
    "linear_strain": "S",
    "tide": "T",
    "wind": "W",
}

measurement_code_dict_reverse = dict(
    [(v, k) for k, v in measurement_code_dict.items()]
)

orientation_code_dict = {
    "N": {"min": 0, "max": 15},
    "E": {"min": 75, "max": 90},
    "Z": {"min": 0, "max": 15},
    "1": {"min": 15, "max": 45},
    "2": {"min": 45, "max": 75},
    "3": {"min": 15, "max": 75},
}

mt_code_dict = {"magnetics": "h", "electric": "e"}


def get_location_code(channel_obj):
    """
    Get the location code given the components and channel number

    :param channel_obj: Channel object
    :type channel_obj: :class:`~mt_metadata.timeseries.Channel`
    :return: 2 character location code
    :rtype: string

    """

    location_code = "{0}{1}".format(
        channel_obj.component[0].upper(),
        channel_obj.channel_number % 10,
    )

    return location_code


def get_period_code(sample_rate):
    """
    Get the SEED sampling rate code given a sample rate

    :param sample_rate: sample rate in samples per second
    :type sample_rate: float
    :return: single character SEED sampling code
    :rtype: string

    """
    period_code = "A"
    for key, v_dict in sorted(period_code_dict.items()):
        if (sample_rate >= v_dict["min"]) and (sample_rate <= v_dict["max"]):
            period_code = key
            break
    return period_code


def get_measurement_code(measurement):
    """
    get SEED sensor code given the measurement type

    :param measurement: measurement type, e.g.
        * temperature
        * electric
        * magnetic
    :type measurement: string
    :return: single character SEED sensor code, if the measurement type has
             not been defined yet Y is returned.
    :rtype: string

    """
    sensor_code = "Y"
    for key, code in measurement_code_dict.items():
        if measurement.lower() in key:
            sensor_code = code
    return sensor_code


def get_orientation_code(azimuth, orientation="horizontal"):
    """
    Get orientation code given angle and orientation.  This is a general
    code and the true azimuth is stored in channel

    :param azimuth: angel assuming 0 is north, 90 is east, 0 is vertical down
    :type azimuth: float
    :return: single character SEED orientation code
    :rtype: string

    """
    orientation_code = "1"
    horizontal_keys = ["N", "E", "1", "2"]
    vertical_keys = ["Z", "3"]

    azimuth = abs(azimuth) % 91
    if orientation == "horizontal":
        test_keys = horizontal_keys

    elif orientation == "vertical":
        test_keys = vertical_keys
    else:
        raise ValueError(
            f"{orientation} not supported must be [ 'horizontal' | 'vertical' ]"
        )

    for key in test_keys:
        angle_min = orientation_code_dict[key]["min"]
        angle_max = orientation_code_dict[key]["max"]
        if (azimuth <= angle_max) and (azimuth >= angle_min):
            orientation_code = key
            break
    return orientation_code


def make_channel_code(channel_obj):
    """
    Make the 3 character SEED channel code

    :param channel_obj: Channel metadata
    :type channel_obj: :class:`~mt_metadata.timeseries.Channel`
    :return: 3 character channel code
    :type: string

    """

    period_code = get_period_code(channel_obj.sample_rate)
    sensor_code = get_measurement_code(channel_obj.type)
    if "z" in channel_obj.component.lower():
        orientation_code = get_orientation_code(
            channel_obj.measurement_tilt, orientation="vertical"
        )
    else:
        orientation_code = get_orientation_code(
            channel_obj.measurement_azimuth
        )

    channel_code = "{0}{1}{2}".format(
        period_code, sensor_code, orientation_code
    )

    return channel_code


def read_channel_code(channel_code):
    """
    read FDSN channel code

    :param channel_code: DESCRIPTION
    :type channel_code: TYPE
    :return: DESCRIPTION
    :rtype: TYPE

    """

    if len(channel_code) != 3:
        msg = (
            "Input FDSN channel code is not proper format, should be 3 letters"
        )
        logger.error(msg)
        raise ValueError(msg)

    try:
        period_range = period_code_dict[channel_code[0].upper()]
    except KeyError:
        msg = (
            f"Could not find period range for {channel_code[0]}. ",
            "Setting to 1",
        )
        period_range = {"min": 1, "max": 1}

    try:
        component = measurement_code_dict_reverse[channel_code[1].upper()]
    except KeyError:
        msg = f"Could not find component for {channel_code[1]}"
        logger.error(msg)
        raise ValueError(msg)

    vertical = False
    try:
        orientation = orientation_code_dict[channel_code[2].upper()]
        if channel_code[2].upper() in ["3", "Z"]:
            vertical = True
    except KeyError:
        msg = (
            f"Could not find orientation for {channel_code[2]}. ",
            "Setting to 0.",
        )
        logger.error(msg)
        raise ValueError(msg)

    return {
        "period": period_range,
        "component": component,
        "orientation": orientation,
        "vertical": vertical,
    }


def make_mt_channel(code_dict, angle_tol=15):
    """

    :param code_dict: DESCRIPTION
    :type code_dict: TYPE
    :return: DESCRIPTION
    :rtype: TYPE

    """

    try:
        mt_comp = mt_code_dict[code_dict["component"]]
    except KeyError:
        mt_comp = code_dict["component"]

    if not code_dict["vertical"]:
        if (
            code_dict["orientation"]["min"] >= 0
            and code_dict["orientation"]["max"] <= angle_tol
        ):
            mt_dir = "x"
        elif (
            code_dict["orientation"]["min"] >= angle_tol
            and code_dict["orientation"]["max"] <= 45
        ):
            mt_dir = "1"
        if (
            code_dict["orientation"]["min"] >= (90 - angle_tol)
            and code_dict["orientation"]["max"] <= 90
        ):
            mt_dir = "y"
        elif code_dict["orientation"]["min"] >= 45 and code_dict[
            "orientation"
        ]["max"] <= (90 - angle_tol):
            mt_dir = "2"

    else:
        if (
            code_dict["orientation"]["min"] >= 0
            and code_dict["orientation"]["max"] <= angle_tol
        ):
            mt_dir = "z"
        elif (
            code_dict["orientation"]["min"] >= angle_tol
            and code_dict["orientation"]["max"] <= 90
        ):
            mt_dir = "3"

    mt_code = f"{mt_comp}{mt_dir}"

    return mt_code



# -*- coding: utf-8 -*-
"""
Created on Sun Sep 11 21:58:30 2022

@author: jpeacock
"""

# =============================================================================
# Imports
# =============================================================================
import unittest

from mth5.utils import fdsn_tools
from mt_metadata.timeseries import Electric, Magnetic, Auxiliary

# =============================================================================


class TestFDSNTools(unittest.TestCase):
    @classmethod
    def setUpClass(self):
        self.ex = Electric(
            component="ex",
            channel_number=1,
            sample_rate=50,
            measurement_azimuth=5,
        )
        self.hx = Magnetic(
            component="hx",
            channel_number=3,
            sample_rate=500,
            measurement_azimuth=95,
        )
        self.tx = Auxiliary(
            component="temperature",
            channel_number=6,
            sample_rate=1200,
            measurement_azimuth=105,
            type="temperature",
        )

    def test_location_code(self):
        with self.subTest("ex"):
            self.assertEqual("E1", fdsn_tools.get_location_code(self.ex))

        with self.subTest("hx"):
            self.assertEqual("H3", fdsn_tools.get_location_code(self.hx))

        with self.subTest("temperature"):
            self.assertEqual("T6", fdsn_tools.get_location_code(self.tx))

    def test_period_code(self):
        code = {
            "F": 2000,
            "C": 500,
            "E": 100,
            "B": 50,
            "M": 5,
            "L": 1,
            "V": 0.1,
            "U": 0.01,
            "R": 0.0005,
            "P": 0.00005,
            "T": 0.000005,
            "Q": 0.0000005,
        }

        for key, value in code.items():
            with self.subTest(key):
                self.assertEqual(key, fdsn_tools.get_period_code(value))

    def test_measurment_code(self):
        code = {
            "tilt": "A",
            "creep": "B",
            "calibration": "C",
            "pressure": "D",
            "magnetics": "F",
            "gravity": "G",
            "humidity": "I",
            "temperature": "K",
            "water_current": "O",
            "electric": "Q",
            "rain_fall": "R",
            "linear_strain": "S",
            "tide": "T",
            "wind": "W",
        }

        for key, value in code.items():
            with self.subTest(msg=key):
                self.assertEqual(value, fdsn_tools.get_measurement_code(key))

    def test_orientation_code_horizontal(self):
        code = {0: "N", 5: "N", 20: "1", 90: "E", 55: "2"}

        for key, value in code.items():
            with self.subTest(msg=str(key)):
                self.assertEqual(value, fdsn_tools.get_orientation_code(key))

    def test_orientation_code_vertical(self):
        code = {0: "Z", 5: "Z", 20: "3"}

        for key, value in code.items():
            with self.subTest(msg=str(key)):
                self.assertEqual(
                    value,
                    fdsn_tools.get_orientation_code(
                        key, orientation="vertical"
                    ),
                )

    def test_orientation_fail(self):
        self.assertRaises(
            ValueError, fdsn_tools.get_orientation_code, 0, "juxtapose"
        )

    def test_make_channel_code(self):
        with self.subTest("ex"):
            self.assertEqual("BQN", fdsn_tools.make_channel_code(self.ex))

        with self.subTest("hx"):
            self.assertEqual("CFN", fdsn_tools.make_channel_code(self.hx))

        with self.subTest("temperature"):
            self.assertEqual("FKN", fdsn_tools.make_channel_code(self.tx))

    def test_read_channel_code(self):
        with self.subTest("BQN"):
            self.assertDictEqual(
                {
                    "period": {"min": 10, "max": 80},
                    "component": "electric",
                    "orientation": {"min": 0, "max": 15},
                    "vertical": False,
                },
                fdsn_tools.read_channel_code("BQN"),
            )

        with self.subTest("FQZ"):
            self.assertDictEqual(
                {
                    "period": {"min": 1000, "max": 5000},
                    "component": "electric",
                    "orientation": {"min": 0, "max": 15},
                    "vertical": True,
                },
                fdsn_tools.read_channel_code("FQZ"),
            )

    def test_make_mt_channel(self):
        with self.subTest("ex"):
            self.assertEqual(
                "ex",
                fdsn_tools.make_mt_channel(
                    fdsn_tools.read_channel_code("BQN")
                ),
            )

        with self.subTest("hx"):
            self.assertEqual(
                "hx",
                fdsn_tools.make_mt_channel(
                    fdsn_tools.read_channel_code("FFN")
                ),
            )

        with self.subTest("temperature"):
            self.assertEqual(
                "temperaturez",
                fdsn_tools.make_mt_channel(
                    fdsn_tools.read_channel_code("KKZ")
                ),
            )


# =============================================================================
# Run
# =============================================================================
if __name__ == "__main__":
    unittest.main()
